# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/servergroupservice/server_group_service__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.auth import service__client_pb2 as tecton__proto_dot_auth_dot_service____client__pb2
from tecton_proto.common import server_group_status__client_pb2 as tecton__proto_dot_common_dot_server__group__status____client__pb2
from tecton_proto.common import server_group_type__client_pb2 as tecton__proto_dot_common_dot_server__group__type____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nBtecton_proto/servergroupservice/server_group_service__client.proto\x12\x1ftecton_proto.servergroupservice\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\'tecton_proto/auth/service__client.proto\x1a\x35tecton_proto/common/server_group_status__client.proto\x1a\x33tecton_proto/common/server_group_type__client.proto\"q\n\x17ListServerGroupsRequest\x12\x1c\n\tworkspace\x18\x01 \x01(\tR\tworkspace\x12\x38\n\x04type\x18\x02 \x01(\x0e\x32$.tecton_proto.common.ServerGroupTypeR\x04type\"q\n\x18ListServerGroupsResponse\x12U\n\rserver_groups\x18\x01 \x03(\x0b\x32\x30.tecton_proto.servergroupservice.ServerGroupInfoR\x0cserverGroups\"a\n\x15GetServerGroupRequest\x12\x1c\n\tworkspace\x18\x01 \x01(\tR\tworkspace\x12*\n\x11server_group_name\x18\x02 \x01(\tR\x0fserverGroupName\"m\n\x16GetServerGroupResponse\x12S\n\x0cserver_group\x18\x01 \x01(\x0b\x32\x30.tecton_proto.servergroupservice.ServerGroupInfoR\x0bserverGroup\"\xfa\x05\n\x0fServerGroupInfo\x12&\n\x0fserver_group_id\x18\x01 \x01(\tR\rserverGroupId\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12\x38\n\x04type\x18\x03 \x01(\x0e\x32$.tecton_proto.common.ServerGroupTypeR\x04type\x12 \n\x0b\x64\x65scription\x18\x04 \x01(\tR\x0b\x64\x65scription\x12\x39\n\ncreated_at\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12\x14\n\x05owner\x18\x06 \x01(\tR\x05owner\x12(\n\x10last_modified_by\x18\x07 \x01(\tR\x0elastModifiedBy\x12N\n\x04tags\x18\x08 \x03(\x0b\x32:.tecton_proto.servergroupservice.ServerGroupInfo.TagsEntryR\x04tags\x12`\n\x0e\x64\x65sired_config\x18\n \x01(\x0b\x32\x39.tecton_proto.servergroupservice.ServerGroupScalingConfigR\rdesiredConfig\x12`\n\x0e\x63urrent_config\x18\x0b \x01(\x0b\x32\x39.tecton_proto.servergroupservice.ServerGroupScalingConfigR\rcurrentConfig\x12>\n\x06status\x18\x0c \x01(\x0e\x32&.tecton_proto.common.ServerGroupStatusR\x06status\x12%\n\x0estatus_details\x18\r \x01(\tR\rstatusDetails\x12 \n\x0b\x65nvironment\x18\x0e \x01(\tR\x0b\x65nvironment\x1a\x37\n\tTagsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"\x9c\x02\n\x18ServerGroupScalingConfig\x12\x1b\n\tmin_nodes\x18\x01 \x01(\x05R\x08minNodes\x12\x1b\n\tmax_nodes\x18\x02 \x01(\x05R\x08maxNodes\x12#\n\rdesired_nodes\x18\x03 \x01(\x05R\x0c\x64\x65siredNodes\x12/\n\x13\x61utoscaling_enabled\x18\x04 \x01(\x08R\x12\x61utoscalingEnabled\x12\x42\n\x0flast_updated_at\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\rlastUpdatedAt\x12,\n\x12workspace_state_id\x18\x06 \x01(\tR\x10workspaceStateId2\x80\x03\n\x12ServerGroupService\x12\xb6\x01\n\x10ListServerGroups\x12\x38.tecton_proto.servergroupservice.ListServerGroupsRequest\x1a\x39.tecton_proto.servergroupservice.ListServerGroupsResponse\"-\xa2\xbc\xe6\xc0\x05\'*%tecton/organization#read_organization\x12\xb0\x01\n\x0eGetServerGroup\x12\x36.tecton_proto.servergroupservice.GetServerGroupRequest\x1a\x37.tecton_proto.servergroupservice.GetServerGroupResponse\"-\xa2\xbc\xe6\xc0\x05\'*%tecton/organization#read_organizationBp\n\x1d\x63om.tecton.servergroupserviceB\x17ServerGroupServiceProtoP\x01Z4github.com/tecton-ai/tecton_proto/servergroupservice')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.servergroupservice.server_group_service__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\035com.tecton.servergroupserviceB\027ServerGroupServiceProtoP\001Z4github.com/tecton-ai/tecton_proto/servergroupservice'
  _SERVERGROUPINFO_TAGSENTRY._options = None
  _SERVERGROUPINFO_TAGSENTRY._serialized_options = b'8\001'
  _SERVERGROUPSERVICE.methods_by_name['ListServerGroups']._options = None
  _SERVERGROUPSERVICE.methods_by_name['ListServerGroups']._serialized_options = b'\242\274\346\300\005\'*%tecton/organization#read_organization'
  _SERVERGROUPSERVICE.methods_by_name['GetServerGroup']._options = None
  _SERVERGROUPSERVICE.methods_by_name['GetServerGroup']._serialized_options = b'\242\274\346\300\005\'*%tecton/organization#read_organization'
  _LISTSERVERGROUPSREQUEST._serialized_start=285
  _LISTSERVERGROUPSREQUEST._serialized_end=398
  _LISTSERVERGROUPSRESPONSE._serialized_start=400
  _LISTSERVERGROUPSRESPONSE._serialized_end=513
  _GETSERVERGROUPREQUEST._serialized_start=515
  _GETSERVERGROUPREQUEST._serialized_end=612
  _GETSERVERGROUPRESPONSE._serialized_start=614
  _GETSERVERGROUPRESPONSE._serialized_end=723
  _SERVERGROUPINFO._serialized_start=726
  _SERVERGROUPINFO._serialized_end=1488
  _SERVERGROUPINFO_TAGSENTRY._serialized_start=1433
  _SERVERGROUPINFO_TAGSENTRY._serialized_end=1488
  _SERVERGROUPSCALINGCONFIG._serialized_start=1491
  _SERVERGROUPSCALINGCONFIG._serialized_end=1775
  _SERVERGROUPSERVICE._serialized_start=1778
  _SERVERGROUPSERVICE._serialized_end=2162
# @@protoc_insertion_point(module_scope)



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Union

DESCRIPTOR: _descriptor.FileDescriptor

class SparkField(_message.Message):
    __slots__ = ["name", "structfield_json"]
    NAME_FIELD_NUMBER: ClassVar[int]
    STRUCTFIELD_JSON_FIELD_NUMBER: ClassVar[int]
    name: str
    structfield_json: str
    def __init__(self, name: Optional[str] = ..., structfield_json: Optional[str] = ...) -> None: ...

class SparkSchema(_message.Message):
    __slots__ = ["fields"]
    FIELDS_FIELD_NUMBER: ClassVar[int]
    fields: _containers.RepeatedCompositeFieldContainer[SparkField]
    def __init__(self, fields: Optional[Iterable[Union[SparkField, Mapping]]] = ...) -> None: ...



import re
from datetime import datetime
from datetime import timezone
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy
import pandas
from pyspark.sql import types as PysparkTypes

from tecton_core import specs
from tecton_core.compute_mode import ComputeMode
from tecton_core.pipeline.rtfv_pipeline import RealtimeFeaturePipeline
from tecton_core.query_consts import udf_internal
from tecton_core.realtime_context import RealtimeContext
from tecton_core.spark_type_annotations import PySparkDataFrame
from tecton_proto.args.pipeline__client_pb2 import Pipeline
from tecton_proto.args.pipeline__client_pb2 import PipelineNode
from tecton_spark.feature_view_spark_utils import check_python_odfv_output_schema


# TODO(TEC-8978): remove \. from namespace regex when FWv3 FVs are no longer supported.
_NAMESPACE_SEPARATOR_REGEX = re.compile(r"__|\.")


def feature_name(namespaced_feature_name: str) -> str:
    """Gets the base feature name from a namespaced_feature_name (e.g. feature_view__feature)

    Supports both `__` (fwv5) and `.` (fwv3) separators. Does two attempts at
    getting the feature name since `__` was allowed in feature view names in
    fwv3.
    """

    spl = _NAMESPACE_SEPARATOR_REGEX.split(namespaced_feature_name)
    if len(spl) == 2:
        return spl[1]

    return namespaced_feature_name.split(".")[1]


def _convert_object_to_serializable_format(item):
    if isinstance(item, numpy.ndarray):
        return [_convert_object_to_serializable_format(value) for value in item.tolist()]
    elif isinstance(item, dict):
        return {key: _convert_object_to_serializable_format(value) for key, value in item.items()}
    elif isinstance(item, list):
        return [_convert_object_to_serializable_format(value) for value in item]
    elif isinstance(item, PysparkTypes.TimestampType):
        return item.fromInternal()
    elif isinstance(item, pandas.Timestamp):
        return item.isoformat()
    elif isinstance(item, pandas.DatetimeIndex):
        return item.isoformat()
    elif isinstance(item, datetime):
        return item.isoformat()
    else:
        return item


class SparkRealtimeFeaturePipeline(RealtimeFeaturePipeline):
    """
    For Pandas-mode:
    A pandas udf takes as inputs List[pandas.Series] and outputs List[pandas.Series]
    However, the user-defined transforms take as input pandas.DataFrame and output
    pandas.DataFrame. RealtimeFeaturePipeline will construct a UDF Wrapper functions
    that translates the inputs and outputs and performs some type checking.

    The general idea is that each Node of the pipeline evaluates to a pandas.DataFrame.
    This is what we want since the user-defined transforms take pandas.DataFrame
    as inputs both from RequestDataSourceNode or FeatureViewNode.
    pandas_udf_wrapper then typechecks and translates the final pandas.DataFrame into a
    jsonized pandas.Series to match what spark expects.

    For Python-mode, we can use a simpler wrapper function for the udf because we don't do
    any spark<->pandas type conversions.
    """

    def __init__(
        self,
        name: str,
        pipeline: Pipeline,
        transformations: List[specs.TransformationSpec],
        # maps input + feature name to arg index that udf function wrapper will be called with.
        # this is needed because we need to know which pandas.Series that are inputs to this
        # function correspond to the desired request context fields or dependent fv features
        # that the customer-defined udf uses.
        udf_arg_idx_map: Dict[str, int],
        output_schema: Optional[PysparkTypes.StructType],
        is_prompt: bool,
        events_df_timestamp_field: Optional[str] = None,
        # the id of this OnDemandFeatureView; only required to be set when reading from source data
        fv_id: Optional[str] = None,
        pipeline_inputs: Optional[Dict[str, Union[Dict[str, Any], pandas.DataFrame, RealtimeContext]]] = None,
    ) -> None:
        self.udf_arg_idx_map = udf_arg_idx_map
        self._output_schema = output_schema
        self._fv_id = fv_id
        self._udf_args = []
        super().__init__(
            name=name,
            pipeline=pipeline,
            transformations=transformations,
            is_prompt=is_prompt,
            events_df_timestamp_field=events_df_timestamp_field,
            pipeline_inputs=pipeline_inputs,
        )

    def _feature_view_node_to_value(
        self, pipeline_node: PipelineNode
    ) -> Union[Dict[str, Any], pandas.DataFrame, PySparkDataFrame]:
        if self._pipeline_inputs is not None:
            return self._pipeline_inputs[pipeline_node.feature_view_node.input_name]
        elif self.is_python_mode:
            fields_dict = {}
            # The input name of this FeatureViewNode tells us which of the udf_args
            # correspond to the Dict we should generate that the parent TransformationNode
            # expects as an input. It also expects the DataFrame to have its columns named
            # by the feature names.
            for feature in self.udf_arg_idx_map:
                if not feature.startswith(
                    f"{udf_internal(ComputeMode.SPARK)}_{pipeline_node.feature_view_node.input_name}_{self._fv_id}"
                ):
                    continue
                idx = self.udf_arg_idx_map[feature]
                value = self._udf_args[idx]
                if isinstance(value, datetime):
                    value = value.replace(tzinfo=timezone.utc)
                fields_dict[feature_name(feature)] = value
            return fields_dict
        elif self.is_pandas_mode:
            all_series = []
            features = []
            # The input name of this FeatureViewNode tells us which of the udf_args
            # correspond to the pandas.DataFrame we should generate that the parent TransformationNode
            # expects as an input. It also expects the DataFrame to have its columns named
            # by the feature names.
            for feature in self.udf_arg_idx_map:
                if not feature.startswith(
                    f"{udf_internal(ComputeMode.SPARK)}_{pipeline_node.feature_view_node.input_name}_{self._fv_id}"
                ):
                    continue
                idx = self.udf_arg_idx_map[feature]
                all_series.append(self._udf_args[idx])
                features.append(feature_name(feature))
            df = pandas.concat(all_series, keys=features, axis=1)
            return self._format_values_for_pandas_mode(df)
        else:
            msg = "Transform mode {self.mode} is not yet implemented"
            raise NotImplementedError(msg)

    def _request_data_node_to_value(
        self, pipeline_node: PipelineNode
    ) -> Union[Dict[str, Any], pandas.DataFrame, PySparkDataFrame]:
        if self._pipeline_inputs is not None:
            return self._pipeline_inputs[pipeline_node.request_data_source_node.input_name]
        elif self.is_python_mode:
            request_context = pipeline_node.request_data_source_node.request_context
            field_names = [c.name for c in request_context.tecton_schema.columns]
            fields_dict = {}
            for input_col in field_names:
                idx = self.udf_arg_idx_map[input_col]
                value = self._udf_args[idx]
                if isinstance(value, datetime):
                    value = value.replace(tzinfo=timezone.utc)
                fields_dict[input_col] = value
            return fields_dict
        elif self.is_pandas_mode:
            all_series = []
            request_context = pipeline_node.request_data_source_node.request_context
            field_names = [c.name for c in request_context.tecton_schema.columns]
            for input_col in field_names:
                idx = self.udf_arg_idx_map[input_col]
                all_series.append(self._udf_args[idx])
            df = pandas.concat(all_series, keys=field_names, axis=1)
            return self._format_values_for_pandas_mode(df)
        else:
            msg = "Transform mode {self.mode} is not yet implemented"
            raise NotImplementedError(msg)

    def _context_node_to_value(self, pipeline_node: PipelineNode) -> RealtimeContext:
        # Used for `run_transformation` with a mock Context
        if self._pipeline_inputs is not None:
            return self._pipeline_inputs[pipeline_node.context_node.input_name]

        if self._events_df_timestamp_field not in self.udf_arg_idx_map:
            msg = f"Could not extract field '{self._events_df_timestamp_field}' from events data frame."
            raise Exception(msg)

        timestamp_index = self.udf_arg_idx_map[self._events_df_timestamp_field]
        request_timestamp = self._udf_args[timestamp_index]

        if self.is_python_mode:
            request_timestamp = request_timestamp.replace(tzinfo=timezone.utc)
            return RealtimeContext(request_timestamp=request_timestamp, _is_python_mode=True)
        elif self.is_pandas_mode:
            df = pandas.to_datetime(request_timestamp, utc=True).to_frame(name="request_timestamp")
            return RealtimeContext(row_level_data=df, _is_python_mode=False)

        else:
            msg = f"{self._fco_name}s must use 'python' or 'pandas' mode."
            raise Exception(msg)

    def py_udf_wrapper(self, *args):
        assert self.is_python_mode
        self._udf_args: List = args
        res = self._node_to_value(self._pipeline.root)
        if self._should_check_output_schema:
            check_python_odfv_output_schema(res, self._output_schema, self._name)
        return res

    def pandas_udf_wrapper(self, *args):
        assert self.is_pandas_mode

        import json

        import pandas

        # self.udf_arg_idx_map tells us which of these pandas.Series correspond to a given
        # RequestDataSource or FeatureView input
        self._udf_args: List[pandas.Series] = args

        output_df = self._node_to_value(self._pipeline.root)
        assert isinstance(
            output_df, pandas.DataFrame
        ), f"Transformer returns {str(output_df)}, but must return a pandas.DataFrame instead."

        for field in self._output_schema:
            assert field.name in output_df.columns, (
                f"Expected output schema field '{field.name}' not found in columns of DataFrame returned by "
                f"'{self._name}': [" + ", ".join(output_df.columns) + "]"
            )
            if isinstance(
                field.dataType,
                (
                    PysparkTypes.ArrayType,
                    PysparkTypes.MapType,
                    PysparkTypes.StructType,
                    PysparkTypes.TimestampType,
                    pandas.Timestamp,
                    pandas.DatetimeIndex,
                ),
            ):
                output_df[field.name] = output_df[field.name].apply(_convert_object_to_serializable_format)

        output_strs = []

        # itertuples() is used instead of iterrows() to preserve type safety.
        # See notes in https://pandas.pydata.org/pandas-docs/version/0.17.1/generated/pandas.DataFrame.iterrows.html.
        for row in output_df.itertuples(index=False):
            output_strs.append(json.dumps(row._asdict()))
        return pandas.Series(output_strs)



from __future__ import annotations

import importlib
from collections import abc
from dataclasses import dataclass
from logging import getLogger
from pathlib import Path
from types import MappingProxyType
from typing import Any
from typing import Callable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import MutableMapping
from typing import Optional

import numpy
import pyarrow
import torch

from tecton_core import data_types
from tecton_core import schema
from tecton_core.embeddings import execution_utils
from tecton_core.errors import TectonInternalError
from tecton_core.errors import TectonValidationError


logger = getLogger(__name__)

_LOAD_FN_NAME = "load_context"
_PRE_PROCESSOR_FN_NAME = "preprocessor"
_POST_PROCESSOR_FN_NAME = "postprocessor"
_RESERVED_MODEL_CONTEXT_KEY = "model"
_TECTON_MODEL_SPEC_NAME = "_tecton_custom_model"

_CONTEXT_PARAM = Mapping[str, Any]
_PRE_PROCESSOR_FN = Optional[Callable[[Mapping[str, numpy.ndarray], _CONTEXT_PARAM], Mapping[str, torch.Tensor]]]
_POST_PROCESSOR_FN = Optional[Callable[[torch.Tensor, _CONTEXT_PARAM], numpy.ndarray]]

_DEFAULT_BATCH_SIZE = 50_000
_MAX_BISECT_RECURSION_LEVEL = 8


def tecton_type_to_pyarrow(tecton_type: data_types.DataType) -> pyarrow.DataType:
    if tecton_type == data_types.Int32Type():
        return pyarrow.int32()
    if tecton_type == data_types.Int64Type():
        return pyarrow.int64()
    if tecton_type == data_types.Float32Type():
        return pyarrow.float32()
    if tecton_type == data_types.Float64Type():
        return pyarrow.float64()
    if tecton_type == data_types.StringType():
        return pyarrow.string()
    if isinstance(tecton_type, data_types.ArrayType):
        return pyarrow.list_(tecton_type_to_pyarrow(tecton_type.element_type))
    msg = f"Invalid tecton type found {tecton_type}"
    raise ValueError(msg)


def _default_pytorch_preprocessor(input: Mapping[str, numpy.ndarray]) -> Mapping[str, torch.Tensor]:
    return {name: torch.tensor(array) for name, array in input.items()}


# TODO(jiadong): CustomModelContainer is pytorch based currently. It can be refactored to a structure that has a `BaseModelContainer` and framework specific model container becomes subclasses of it. The base class can have `load` function implmeneted which is framework agonistic and each subclass implements the `predict` function.
class CustomModelContainer:
    _data_dir: Path
    _model_file_path: str

    # Late init params
    _context: MutableMapping[str, Any]

    _preprocessor: _PRE_PROCESSOR_FN = None
    _postprocessor: _POST_PROCESSOR_FN = None

    def __init__(self, data_dir: str, model_file_path: str = "model.py") -> None:
        self._data_dir = Path(data_dir)
        self._model_file_path = model_file_path
        self._context = {}

    def load(self):
        model_module = self._load_model_module()

        _load_fn = getattr(model_module, _LOAD_FN_NAME, None)
        if not _load_fn:
            msg = f"`{_LOAD_FN_NAME}` function is missing in the model repo"
            raise AttributeError(msg)

        _load_fn(self._data_dir, self._context)

        model = self._context.get(_RESERVED_MODEL_CONTEXT_KEY, None)
        if not model:
            msg = "No 'model' found in the context. `load` function should initialize the model and put it in the context with 'model' key."
            raise ValueError(msg)

        self._preprocessor = getattr(model_module, _PRE_PROCESSOR_FN_NAME, None)
        self._postprocessor = getattr(model_module, _POST_PROCESSOR_FN_NAME, None)

    def _load_model_module(self):
        spec = importlib.util.spec_from_file_location(_TECTON_MODEL_SPEC_NAME, self._data_dir / self._model_file_path)
        model_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(model_module)
        return model_module

    def predict(self, input: pyarrow.RecordBatch) -> numpy.ndarray:
        # zero_copy_only=False is needed to allow conversion of non-primitive types see more:
        # https://arrow.apache.org/docs/python/generated/pyarrow.Int16Array.html#pyarrow.Int16Array.to_numpy
        numpy_dict = {
            name: column.to_numpy(zero_copy_only=False) for name, column in zip(input.schema.names, input.columns)
        }
        if self._preprocessor:
            input = self._preprocessor(numpy_dict, self._context)
        else:
            input = _default_pytorch_preprocessor(numpy_dict)

        if not isinstance(input, abc.Mapping):
            msg = f"`preprocessor` needs to return a 'Mapping[str, torch.Tensor]', but received '{type(input)}'"
            raise TypeError(msg)

        for k, v in input.items():
            if not isinstance(v, torch.Tensor):
                msg = f"`preprocessor` needs to return a 'Mapping[str, torch.Tensor]' but detected '{k}' returned by preprocessor is an instance of type {type(v)}."
                raise TypeError(msg)

        model = self._context[_RESERVED_MODEL_CONTEXT_KEY]
        with torch.no_grad():
            output = model(**input)

        if isinstance(output, torch.Tensor):
            output = output.cpu()

        if self._postprocessor:
            output = self._postprocessor(output, self._context)

            if not isinstance(output, numpy.ndarray):
                msg = (
                    f"`postprocessor` needs to return a 'numpy.ndarray' but detected an instance of type {type(output)}"
                )
                raise TypeError(msg)
        else:
            output = output.numpy()

        return output


# TODO(EMBED-117): Bisect input data if the inference OOM.
def custom_model_inference(
    batch: pyarrow.RecordBatch,
    *,
    model: CustomModelContainer,
    input_columns: List[schema.Column],
    output_column: schema.Column,
    cuda_device: str,
    model_input_schema: List[schema.Column],
) -> pyarrow.RecordBatch:
    with torch.device(cuda_device):
        return _custom_model_inference(batch, model, input_columns, output_column, model_input_schema)


def _custom_model_inference(
    batch: pyarrow.RecordBatch,
    model: CustomModelContainer,
    input_columns: List[schema.Column],
    output_column: schema.Column,
    model_input_schema: List[schema.Column],
) -> pyarrow.RecordBatch:
    input_batch = batch.select([col.name for col in input_columns])
    model_input_names = [col.name for col in model_input_schema]

    # Rename input columns to model input schema column names.
    # TODO(jiadong): we can use pyarrow built-in `rename_columns` when we upgrade pyarrow to 16.0+.
    input_columns = [input_batch.column(i) for i in range(input_batch.num_columns)]
    input_batch = pyarrow.RecordBatch.from_arrays(input_columns, model_input_names)

    output = _custom_model_predict(model=model, input_batch=input_batch)
    output = _convert_pytorch_to_arrow(output, batch.num_rows, tecton_type_to_pyarrow(output_column.dtype))

    batch_with_results = pyarrow.RecordBatch.from_arrays(
        [*batch.columns, output],
        schema=batch.schema.append(pyarrow.field(output_column.name, output.type)),
    )
    return batch_with_results


def _custom_model_predict(
    model: CustomModelContainer, input_batch: pyarrow.RecordBatch, recursion_level: int = 0
) -> numpy.ndarray:
    try:
        return model.predict(input_batch)
    except torch.cuda.OutOfMemoryError as oom_exc:
        logger.warning(torch.cuda.memory_summary())
        logger.warning(f"Out of Memory. {oom_exc}\n")
        execution_utils.gc_all(torch.cuda.current_device())
    return _custom_model_predict_with_bisect(model=model, input_batch=input_batch, recursion_level=recursion_level + 1)


def _custom_model_predict_with_bisect(
    model: CustomModelContainer, input_batch: pyarrow.RecordBatch, recursion_level: int
) -> numpy.ndarray:
    if recursion_level == _MAX_BISECT_RECURSION_LEVEL:
        msg = "Inference is out of memory and out of retries."
        raise TectonInternalError(msg)

    logger.warning(
        f"Attempt {recursion_level}/{_MAX_BISECT_RECURSION_LEVEL} to bisect the input batch and trying again."
    )
    input_size = input_batch.num_rows
    mid_point = input_size // 2

    first_half = input_batch.slice(offset=0, length=mid_point)
    second_half = input_batch.slice(offset=mid_point)
    return numpy.concatenate(
        [
            _custom_model_predict(model, first_half, recursion_level),
            _custom_model_predict(model, second_half, recursion_level),
        ]
    )


def _convert_pytorch_to_arrow(
    output: numpy.ndarray, num_of_rows: int, output_column_type: pyarrow.DataType
) -> pyarrow.Array:
    # If the output is scalar but is 2d (n x 1), then conversion to pyArrow will fail with scalar output_column_type
    # so we need to remove that extra dimension
    if output.ndim == 2 and not isinstance(output_column_type, pyarrow.ListType):
        output = output.squeeze()

    if output.shape[0] != num_of_rows:
        msg = "Output of model should have a single row per input row. "
        raise TectonValidationError(msg)
    return pyarrow.array(list(output), type=output_column_type)


# TODO(jiadong): Implement custom model batch function.
def default_custom_batch_fn(
    batch: pyarrow.RecordBatch,
) -> Iterator[pyarrow.RecordBatch]:
    index = 0
    while index < batch.num_rows:
        num_rows = min(_DEFAULT_BATCH_SIZE, batch.num_rows - index)
        yield batch.take(list(range(index, index + num_rows)))
        index += num_rows


@dataclass
class CustomModelInferenceFuncConfig:
    _model_dir: str
    _model_file_path: str
    _extra_kwargs: Mapping[str, Any]
    _cuda_device: str

    # Note: _final_kwargs is a late init object.
    _final_kwargs: Optional[Mapping[str, Any]] = None

    @classmethod
    def create(
        cls,
        model_dir: str,
        model_file_path: str,
        cuda_device: str,
        input_columns: List[schema.Column],
        output_column: schema.Column,
        model_input_schema: List[schema.Column],
    ) -> CustomModelInferenceFuncConfig:
        return cls(
            _model_dir=model_dir,
            _model_file_path=model_file_path,
            _cuda_device=cuda_device,
            _extra_kwargs=MappingProxyType(
                {
                    "cuda_device": cuda_device,
                    "input_columns": input_columns,
                    "output_column": output_column,
                    "model_input_schema": model_input_schema,
                }
            ),
        )

    def load(self):
        model_container = CustomModelContainer(data_dir=self._model_dir, model_file_path=self._model_file_path)
        with torch.device(self._cuda_device):
            model_container.load()

        self._final_kwargs = MappingProxyType(dict(model=model_container, **self._extra_kwargs))
        return self

    def kwargs(self) -> Mapping[str, Any]:
        if self._final_kwargs is None:
            msg = "`load` must be called prior to calling `kwargs`."
            raise ValueError(msg)
        return self._final_kwargs


@dataclass
class CustomModelBatchFuncConfig:
    # TODO(jiadong): Add necessary params to batch function config such as user-provided batch size.

    # Note: _final_kwargs is a late init object.
    _final_kwargs: Optional[Mapping[str, Any]] = None

    @classmethod
    def create(cls) -> CustomModelBatchFuncConfig:
        return cls()

    def load(self):
        self._final_kwargs = MappingProxyType({})
        return self

    def kwargs(self) -> Mapping[str, Any]:
        if self._final_kwargs is None:
            msg = "`load` must be called prior to calling `kwargs`."
            raise ValueError(msg)
        return self._final_kwargs



from tecton_proto.realtime import instance_group__client_pb2 as _instance_group__client_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Mapping, Optional, Union

DESCRIPTOR: _descriptor.FileDescriptor

class ColocatedComputeConfig(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class InstanceGroupComputeConfig(_message.Message):
    __slots__ = ["group_name", "instance_group"]
    GROUP_NAME_FIELD_NUMBER: ClassVar[int]
    INSTANCE_GROUP_FIELD_NUMBER: ClassVar[int]
    group_name: str
    instance_group: _instance_group__client_pb2.InstanceGroupHandle
    def __init__(self, group_name: Optional[str] = ..., instance_group: Optional[Union[_instance_group__client_pb2.InstanceGroupHandle, Mapping]] = ...) -> None: ...

class OnlineComputeConfig(_message.Message):
    __slots__ = ["colocated_compute", "instance_group_config", "remote_compute"]
    COLOCATED_COMPUTE_FIELD_NUMBER: ClassVar[int]
    INSTANCE_GROUP_CONFIG_FIELD_NUMBER: ClassVar[int]
    REMOTE_COMPUTE_FIELD_NUMBER: ClassVar[int]
    colocated_compute: ColocatedComputeConfig
    instance_group_config: InstanceGroupComputeConfig
    remote_compute: RemoteFunctionComputeConfig
    def __init__(self, colocated_compute: Optional[Union[ColocatedComputeConfig, Mapping]] = ..., remote_compute: Optional[Union[RemoteFunctionComputeConfig, Mapping]] = ..., instance_group_config: Optional[Union[InstanceGroupComputeConfig, Mapping]] = ...) -> None: ...

class RemoteFunctionComputeConfig(_message.Message):
    __slots__ = ["function_uri", "id", "name"]
    FUNCTION_URI_FIELD_NUMBER: ClassVar[int]
    ID_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    function_uri: str
    id: str
    name: str
    def __init__(self, id: Optional[str] = ..., name: Optional[str] = ..., function_uri: Optional[str] = ...) -> None: ...



import logging
from datetime import datetime
from datetime import timedelta
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import pandas
from pyspark import sql as pyspark_sql

import tecton_core.query.dialect
import tecton_core.tecton_pendulum as pendulum
from tecton import types as sdk_types
from tecton._internals import errors
from tecton._internals import type_utils
from tecton._internals import utils
from tecton.framework.data_frame import TectonDataFrame
from tecton_core import conf
from tecton_core import data_types
from tecton_core import errors as core_errors
from tecton_core import feature_set_config
from tecton_core import schema
from tecton_core import specs
from tecton_core import time_utils
from tecton_core.compute_mode import ComputeMode
from tecton_core.data_processing_utils import should_infer_timestamp_of_spine
from tecton_core.data_processing_utils import split_range
from tecton_core.data_processing_utils import split_spine
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.feature_set_config import FeatureDefinitionAndJoinConfig
from tecton_core.query import builder
from tecton_core.query import node_interface
from tecton_core.query import nodes
from tecton_core.query import rewrite
from tecton_core.query.builder import build_materialization_querytree
from tecton_core.query.builder import build_pipeline_querytree
from tecton_core.query.dialect import Dialect
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.retrieval_params import GetFeaturesForEventsParams
from tecton_core.query.retrieval_params import GetFeaturesInRangeParams
from tecton_core.query.rewrite import MockDataRewrite
from tecton_core.schema import Schema
from tecton_core.time_utils import align_time_downwards
from tecton_core.time_utils import temporal_fv_get_feature_data_time_limits
from tecton_proto.common import schema__client_pb2 as schema_pb2
from tecton_spark.spark_helper import check_spark_version
from tecton_spark.time_utils import convert_epoch_to_datetime
from tecton_spark.time_utils import convert_timestamp_to_epoch


logger = logging.getLogger(__name__)


def get_features_from_params(
    params: Union[GetFeaturesInRangeParams, GetFeaturesForEventsParams],
):
    if isinstance(params, GetFeaturesForEventsParams):
        if isinstance(params.fco, specs.FeatureServiceSpec):
            if conf.get_bool("DUCKDB_ENABLE_SPINE_SPLIT") and params.compute_mode == ComputeMode.RIFT:
                df = get_historical_features_for_feature_service_with_spine_split(
                    dialect=params.compute_mode.default_dialect(),
                    compute_mode=params.compute_mode,
                    feature_set_config=params.feature_set_config,
                    spine=params.events,
                    timestamp_key=params.timestamp_key,
                    from_source=params.from_source,
                )
            else:
                df = get_historical_features_for_feature_service(
                    dialect=params.compute_mode.default_dialect(),
                    compute_mode=params.compute_mode,
                    feature_set_config=params.feature_set_config,
                    spine=params.events,
                    timestamp_key=params.timestamp_key,
                    from_source=params.from_source,
                )
        elif isinstance(params.fco, FeatureDefinitionWrapper):
            df = get_features_for_events(
                dialect=params.compute_mode.default_dialect(),
                compute_mode=params.compute_mode,
                feature_definition=params.fco,
                spine=params.events,
                timestamp_key=params.timestamp_key,
                from_source=params.from_source,
                mock_data_sources=params.mock_data_sources,
            )
    elif isinstance(params, GetFeaturesInRangeParams):
        if conf.get_bool("DUCKDB_ENABLE_RANGE_SPLIT") and params.compute_mode == ComputeMode.RIFT:
            ranges = split_range(params.start_time, params.end_time, params.fco.min_scheduling_interval)
            dfs = []
            for start, end in ranges:
                df = get_features_in_range(
                    dialect=params.compute_mode.default_dialect(),
                    compute_mode=params.compute_mode,
                    feature_definition=params.fco,
                    start_time=start,
                    end_time=end,
                    max_lookback=params.max_lookback,
                    entities=params.entities,
                    from_source=params.from_source,
                    mock_data_sources=params.mock_data_sources,
                    valid_from=params.start_time,
                    valid_to=params.end_time,
                )
                dfs.append(df)
            df = TectonDataFrame._create_from_dataframes(dfs)
        else:
            df = get_features_in_range(
                dialect=params.compute_mode.default_dialect(),
                compute_mode=params.compute_mode,
                feature_definition=params.fco,
                start_time=params.start_time,
                end_time=params.end_time,
                max_lookback=params.max_lookback,
                entities=params.entities,
                from_source=params.from_source,
                mock_data_sources=params.mock_data_sources,
            )
    else:
        error = f"Unsupported params type: {type(params)}"
        raise ValueError(error)

    df._request_params = params
    return df


def get_historical_features_for_feature_service_with_spine_split(
    dialect: Dialect,
    compute_mode: ComputeMode,
    feature_set_config: feature_set_config.FeatureSetConfig,
    spine: Union[pyspark_sql.DataFrame, pandas.DataFrame, TectonDataFrame, str],
    timestamp_key: Optional[str],
    from_source: Optional[bool],
) -> TectonDataFrame:
    spine_split = split_spine(spine, feature_set_config.join_keys)
    dfs = []
    for spine_df in spine_split:
        df = get_historical_features_for_feature_service(
            dialect=dialect,
            compute_mode=compute_mode,
            feature_set_config=feature_set_config,
            spine=spine_df,
            timestamp_key=timestamp_key,
            from_source=from_source,
        )
        dfs.append(df)
    return TectonDataFrame._create_from_dataframes(dfs)


def get_historical_features_for_feature_service(
    dialect: Dialect,
    compute_mode: ComputeMode,
    feature_set_config: feature_set_config.FeatureSetConfig,
    spine: Union[pyspark_sql.DataFrame, pandas.DataFrame, TectonDataFrame, str],
    timestamp_key: Optional[str],
    from_source: Optional[bool],
) -> TectonDataFrame:
    timestamp_required = spine is not None and any(
        should_infer_timestamp_of_spine(fd, timestamp_key) for fd in feature_set_config.feature_definitions
    )

    _validate_sql_string_support(dialect, spine)

    if timestamp_required:
        timestamp_key = timestamp_key or utils.infer_timestamp(spine)

    spine = _create_tecton_df_spine(
        spine_schema=feature_set_config.spine_schema, spine=spine, timestamp_key=timestamp_key
    )

    if spine:
        utils.validate_spine_dataframe(spine, timestamp_key, feature_set_config.request_context_keys)

    user_data_node_metadata = {}
    # TODO: Create a SpineNode with a param of timestamp_key instead of using UserSpecifiedNode.
    user_data_node = nodes.UserSpecifiedDataNode(dialect, compute_mode, spine, user_data_node_metadata)

    if timestamp_key:
        user_data_node_metadata["timestamp_key"] = timestamp_key
        user_data_node = nodes.ConvertTimestampToUTCNode(dialect, compute_mode, user_data_node.as_ref(), timestamp_key)

    tree = builder.build_feature_set_config_querytree(
        dialect,
        compute_mode,
        feature_set_config,
        user_data_node.as_ref(),
        timestamp_key,
        from_source,
    )

    df = TectonDataFrame._create(tree)
    return df


def get_features_in_range(
    dialect: Dialect,
    compute_mode: ComputeMode,
    feature_definition: FeatureDefinitionWrapper,
    start_time: Union[pendulum.DateTime, datetime],
    end_time: Union[pendulum.DateTime, datetime],
    max_lookback: Optional[timedelta],
    entities: Optional[Union[pyspark_sql.DataFrame, pandas.DataFrame, TectonDataFrame]],
    from_source: Optional[bool],
    mock_data_sources: Dict[str, NodeRef],
    valid_from: Optional[Union[pendulum.DateTime, datetime]] = None,
    valid_to: Optional[Union[pendulum.DateTime, datetime]] = None,
) -> TectonDataFrame:
    """
    Returns a TectonDataFrame of historical values for this feature view which were valid within
    the input time range. A feature value is considered to be valid at a specific point in time if the Online Store
    would have returned that value if queried at that moment in time.

    The DataFrame returned by this method contains the following:
    - Entity Join Key Columns
    - Feature Value Columns
    - The columns _valid_from and _valid_to that specify the time range for which the row of features values is
    valid. The time range defined by [_valid_from, _valid_to) will never intersect with any other rows for the same
    join keys.

    _valid_from (Inclusive) is the timestamp from which feature values were valid and returned from the Online Feature
    Store for the corresponding set of join keys. _valid_from will never be less than end_time.
    Values for which _valid_from is equal to start_time may have been valid prior to start_time.

    _valid_to (Exclusive) is the timestamp from which feature values are invalid and no longer returned from the
    Online Feature Store for the corresponding set of join keys. _valid_to will never be greater than end_time.
    Values for which _valid_to is equal to end_time may be valid beyond end_time.

    By default, (i.e. from_source=None), this method fetches feature values from the Offline Store for
    Feature Views that have offline materialization enabled. Otherwise, this method computes feature values directly
    from the original data source.

    :param start_time: The interval start time from when we want to retrieve features.
    :param end_time:  The interval end time until when we want to retrieve features.
    :param max_lookback: [Non-Aggregate Feature Views Only] A performance optimization that configures how far back
    before start_time to look for events in the raw data. If set, get_features_in_range() may not include all
    entities with valid feature values in the specified time range, but get_features_in_range() will never
    return invalid values.
    :param entities: A DataFrame that is used to filter down feature values.
        If specified, this DataFrame should only contain join key columns.
    :param from_source: Whether feature values should be recomputed from the original data source. If None,
    feature values will be fetched from the Offline Store for Feature Views that have offline materialization enabled
    and otherwise computes feature values on the fly from raw data. Use from_source=True to force computing from raw
    data and from_source=False to error if any Feature Views are not materialized. (Default: None)
    :param compute_mode: Compute mode to use to produce the data frame.
    :param mock_data_sources: Mock sources that should be used instead of fetching directly from raw data
        sources. The keys of the dictionary should match the feature view's function parameters. For feature views with multiple sources, mocking some data sources and using raw data for others is supported.
    :param valid_from: The start time for which the feature values are valid. If not provided, defaults to start_time.
    :param valid_to: The end time for which the feature values are valid. If not provided, defaults to end_time.

    :return: A TectonDataFrame with Feature Values for the requested time range in the format specified above.
    """
    if not feature_definition.is_rtfv_or_prompt:
        check_spark_version(feature_definition.fv_spec.batch_cluster_config)

    if entities is not None:
        if not isinstance(entities, TectonDataFrame):
            entities = TectonDataFrame._create(entities)
        assert set(entities._dataframe.columns).issubset(
            set(feature_definition.join_keys)
        ), f"Entities should only contain columns that can be used as Join Keys: {feature_definition.join_keys}"

    start_time = pendulum.instance(start_time)
    end_time = pendulum.instance(end_time)
    if feature_definition.feature_start_timestamp is not None:
        start_time = max(start_time, feature_definition.feature_start_timestamp)
    query_time_range = pendulum.Period(start_time, end_time)
    valid_from = valid_from or start_time
    valid_to = valid_to or end_time
    # The bounds of the time range for which the feature values are valid, valid_to and valid_from values will be capped by the valid_time_range
    valid_time_range = pendulum.Period(valid_from, valid_to)

    if feature_definition.is_temporal or feature_definition.is_feature_table:
        lookback_time_range = temporal_fv_get_feature_data_time_limits(
            feature_definition, query_time_range, max_lookback
        )
        qt = builder.build_temporal_time_range_validity_query(
            dialect=dialect,
            compute_mode=compute_mode,
            fd=feature_definition,
            from_source=from_source,
            lookback_time_range=lookback_time_range,
            entities=entities,
            query_time_range=valid_time_range,
        )
    else:
        feature_data_time_limits = time_utils.get_feature_data_time_limits(
            fd=feature_definition,
            spine_time_limits=query_time_range,
        )
        qt = builder.build_aggregated_time_range_validity_query(
            dialect,
            compute_mode,
            feature_definition,
            feature_data_time_limits=feature_data_time_limits,
            query_time_range=query_time_range,
            from_source=from_source,
            entities=entities,
        )

    rewrite.MockDataRewrite(mock_data_sources).rewrite(qt)

    df = TectonDataFrame._create(qt)

    return df


def get_partial_aggregates(
    dialect: Dialect,
    compute_mode: ComputeMode,
    feature_definition: FeatureDefinitionWrapper,
    start_time: Union[pendulum.DateTime, datetime],
    end_time: Union[pendulum.DateTime, datetime],
    entities: Optional[TectonDataFrame],
    from_source: Optional[bool],
    mock_data_sources: Dict[str, NodeRef],
) -> TectonDataFrame:
    check_spark_version(feature_definition.fv_spec.batch_cluster_config)

    start_time = pendulum.instance(start_time)
    if feature_definition.feature_start_timestamp is not None:
        start_time = max(start_time, feature_definition.feature_start_timestamp)

    start_time = align_time_downwards(start_time, feature_definition.min_scheduling_interval)
    end_time = (
        align_time_downwards(end_time, feature_definition.min_scheduling_interval)
        + feature_definition.min_scheduling_interval
    )
    limits = pendulum.Period(start_time, end_time)

    qt = builder.build_get_partial_aggregates_query(
        dialect,
        compute_mode,
        feature_definition,
        limits=limits,
        from_source=from_source,
        entities=entities,
    )

    rewrite.MockDataRewrite(mock_data_sources).rewrite(qt)

    df = TectonDataFrame._create(qt)

    return df


def get_features_for_events(
    dialect: Dialect,
    compute_mode: ComputeMode,
    feature_definition: FeatureDefinitionWrapper,
    spine: Optional[Union[pyspark_sql.DataFrame, pandas.DataFrame, TectonDataFrame, str]],
    timestamp_key: Optional[str],
    from_source: Optional[bool],
    mock_data_sources: Dict[str, NodeRef],
):
    if not feature_definition.is_rtfv_or_prompt:
        check_spark_version(feature_definition.fv_spec.batch_cluster_config)

    _validate_sql_string_support(dialect, spine)

    if should_infer_timestamp_of_spine(feature_definition, timestamp_key):
        timestamp_key = utils.infer_timestamp(spine)

    if conf.get_bool("DUCKDB_ENABLE_SPINE_SPLIT") and compute_mode == ComputeMode.RIFT:
        spine_split = split_spine(spine, feature_definition.join_keys)
        dfs = []
        for spine_df in spine_split:
            tecton_spine_df = _create_tecton_df_spine(
                spine_schema=feature_definition.spine_schema, spine=spine_df, timestamp_key=timestamp_key
            )

            qt = _point_in_time_get_historical_features_for_feature_definition(
                dialect, compute_mode, feature_definition, tecton_spine_df, timestamp_key, from_source
            )

            rewrite.MockDataRewrite(mock_data_sources).rewrite(qt)

            df = TectonDataFrame._create(qt)
            dfs.append(df)

        return TectonDataFrame._create_from_dataframes(dfs)

    tecton_spine_df = _create_tecton_df_spine(
        spine_schema=feature_definition.spine_schema, spine=spine, timestamp_key=timestamp_key
    )

    qt = _point_in_time_get_historical_features_for_feature_definition(
        dialect, compute_mode, feature_definition, tecton_spine_df, timestamp_key, from_source
    )

    rewrite.MockDataRewrite(mock_data_sources).rewrite(qt)

    df = TectonDataFrame._create(qt)
    return df


def get_historical_features_for_feature_definition(
    dialect: Dialect,
    compute_mode: ComputeMode,
    feature_definition: FeatureDefinitionWrapper,
    spine: Optional[Union[pyspark_sql.DataFrame, pandas.DataFrame, TectonDataFrame, str]],
    timestamp_key: Optional[str],
    start_time: Optional[Union[pendulum.DateTime, datetime]],
    end_time: Optional[Union[pendulum.DateTime, datetime]],
    entities: Optional[Union[pyspark_sql.DataFrame, pandas.DataFrame, TectonDataFrame]],
    from_source: Optional[bool],
    mock_data_sources: Dict[str, NodeRef],
) -> TectonDataFrame:
    if not feature_definition.is_rtfv_or_prompt:
        check_spark_version(feature_definition.fv_spec.batch_cluster_config)

    if spine is not None:
        logger.warning(errors.GET_HISTORICAL_FEATURES_DEPRECATED_SPINE)

        _validate_sql_string_support(dialect, spine)

        if should_infer_timestamp_of_spine(feature_definition, timestamp_key):
            timestamp_key = utils.infer_timestamp(spine)

        tecton_spine_df = _create_tecton_df_spine(
            spine_schema=feature_definition.spine_schema, spine=spine, timestamp_key=timestamp_key
        )
        qt = _point_in_time_get_historical_features_for_feature_definition(
            dialect, compute_mode, feature_definition, tecton_spine_df, timestamp_key, from_source
        )
    else:
        logger.warning(errors.GET_HISTORICAL_FEATURES_DEPRECATED_TIME_RANGE)

        if entities is not None:
            if not isinstance(entities, TectonDataFrame):
                entities = TectonDataFrame._create(entities)
            assert set(entities._dataframe.columns).issubset(
                set(feature_definition.join_keys)
            ), f"Entities should only contain columns that can be used as Join Keys: {feature_definition.join_keys}"

        query_time_range = _get_feature_data_time_range(feature_definition, start_time, end_time)
        tecton_spine_df = None

        if feature_definition.is_temporal or feature_definition.is_feature_table:
            qt = builder.build_temporal_time_range_query(
                dialect=dialect,
                compute_mode=compute_mode,
                fd=feature_definition,
                from_source=from_source,
                query_time_range=query_time_range,
                entities=entities,
            )
        else:
            feature_data_time_limits = time_utils.get_feature_data_time_limits(
                fd=feature_definition,
                spine_time_limits=query_time_range,
            )
            qt = builder.build_aggregated_time_range_ghf_query(
                dialect,
                compute_mode,
                feature_definition,
                from_source=from_source,
                feature_data_time_limits=feature_data_time_limits,
                query_time_range=query_time_range,
                entities=entities,
            )

    rewrite.MockDataRewrite(mock_data_sources).rewrite(qt)

    df = TectonDataFrame._create(qt)
    return df


def _validate_sql_string_support(dialect: Dialect, spine):
    """
    Checks that the spine is a supported type for the given dialect.
    """
    if isinstance(spine, str) and dialect not in (
        tecton_core.query.dialect.Dialect.SNOWFLAKE,
        tecton_core.query.dialect.Dialect.DUCKDB,
    ):
        msg = "Using SQL str as `spine` is only supported with Snowflake."
        raise TypeError(msg)


def _point_in_time_get_historical_features_for_feature_definition(
    dialect: Dialect,
    compute_mode: ComputeMode,
    feature_definition: FeatureDefinitionWrapper,
    spine: TectonDataFrame,
    timestamp_key: Optional[str],
    from_source: Optional[bool],
) -> node_interface.NodeRef:
    if feature_definition.is_rtfv_or_prompt:
        utils.validate_spine_dataframe(spine, timestamp_key, feature_definition.request_context_keys)
    else:
        utils.validate_spine_dataframe(spine, timestamp_key)

    dac = FeatureDefinitionAndJoinConfig.from_feature_definition(feature_definition)
    user_data_node_metadata = {}
    user_data_node = nodes.UserSpecifiedDataNode(dialect, compute_mode, spine, user_data_node_metadata)

    if timestamp_key:
        user_data_node_metadata["timestamp_key"] = timestamp_key
        user_data_node = nodes.ConvertTimestampToUTCNode(dialect, compute_mode, user_data_node.as_ref(), timestamp_key)

    qt = builder.build_spine_join_querytree(
        dialect,
        compute_mode,
        dac,
        user_data_node.as_ref(),
        timestamp_key,
        from_source,
        use_namespace_feature_prefix=dialect != tecton_core.query.dialect.Dialect.SNOWFLAKE,
    )

    return qt


def _most_recent_tile_end_time(fd: FeatureDefinitionWrapper, timestamp: datetime) -> int:
    """Computes the most recent tile end time which is ready to be computed.

    :param timestamp: The timestamp in python datetime format
    :return: The timestamp in seconds of the greatest ready tile end time <= timestamp.
    """
    # Account for data delay
    timestamp = timestamp - fd.max_source_data_delay
    if fd.min_scheduling_interval:
        timestamp = align_time_downwards(timestamp, fd.min_scheduling_interval)
    return convert_timestamp_to_epoch(timestamp, fd.get_feature_store_format_version)


def _get_feature_data_time_range(
    fd: FeatureDefinitionWrapper,
    start_time: Optional[Union[pendulum.DateTime, datetime]],
    end_time: Optional[Union[pendulum.DateTime, datetime]],
) -> pendulum.Period:
    if start_time is not None and isinstance(start_time, datetime):
        start_time = pendulum.instance(start_time)

    if end_time is not None and isinstance(end_time, datetime):
        end_time = pendulum.instance(end_time)

    if start_time is not None and fd.feature_start_timestamp is not None and start_time < fd.feature_start_timestamp:
        logger.warning(
            f'The provided start_time ({start_time}) is before "{fd.name}"\'s feature_start_time ({fd.feature_start_timestamp}). No feature values will be returned before the feature_start_time.'
        )
        start_time = fd.feature_start_timestamp

    # TODO(brian): construct the timestamps a bit more directly. This code in
    # general reuses utilities not really meant for the semantics of this API.
    if fd.is_temporal_aggregate or fd.is_temporal:
        # Feature views where materialization is not enabled may not have a feature_start_time.
        _start = start_time or fd.feature_start_timestamp or pendulum.datetime(1970, 1, 1)
        # we need to add 1 to most_recent_anchor since we filter end_time exclusively
        if end_time:
            _end = end_time
        else:
            anchor_time = _most_recent_tile_end_time(fd, pendulum.now("UTC") - fd.min_scheduling_interval)
            _end = convert_epoch_to_datetime(anchor_time, fd.get_feature_store_format_version) + pendulum.duration(
                microseconds=1
            )
    else:
        _start = start_time or pendulum.datetime(1970, 1, 1)
        _end = end_time or pendulum.now("UTC")

    if _start >= _end:
        # TODO(felix): Move this and other instances of validating user inputs to top-level get_historical_features() methods.
        raise core_errors.START_TIME_NOT_BEFORE_END_TIME(_start, _end)

    return pendulum.Period(_start, _end)


def get_fields_list_from_tecton_schema(tecton_schema: schema_pb2.Schema) -> List[sdk_types.Field]:
    """Convert TectonSchema into a list of Tecton Fields."""
    columns_and_types = schema.Schema(tecton_schema).column_name_and_data_types()
    request_schema = []
    for c_and_t in columns_and_types:
        name = c_and_t[0]
        data_type = type_utils.sdk_type_from_tecton_type(c_and_t[1])
        request_schema.append(sdk_types.Field(name, data_type))
    return request_schema


def _create_tecton_df_spine(
    spine_schema: Schema,
    spine: Optional[Union[pyspark_sql.DataFrame, pandas.DataFrame, TectonDataFrame, str]],
    timestamp_key: Optional[str],
) -> TectonDataFrame:
    if isinstance(spine, pandas.DataFrame):
        if timestamp_key is not None:
            spine_schema += schema.Schema.from_dict({timestamp_key: data_types.TimestampType()})
        spine = TectonDataFrame._create_from_pandas_with_schema(spine, schema=spine_schema)
    elif isinstance(spine, str):
        spine = TectonDataFrame._create_with_snowflake_sql(spine)
    elif not isinstance(spine, TectonDataFrame):
        spine = TectonDataFrame._create(spine)

    return spine


def get_dataframe_for_data_source(
    dialect: Dialect,
    compute_mode: ComputeMode,
    data_source: specs.DataSourceSpec,
    start_time: Optional[datetime],
    end_time: Optional[datetime],
    apply_translator: bool,
) -> TectonDataFrame:
    if isinstance(data_source.batch_source, (specs.SparkBatchSourceSpec, specs.PandasBatchSourceSpec)):
        if not data_source.batch_source.supports_time_filtering and (start_time or end_time):
            raise errors.DS_INCORRECT_SUPPORTS_TIME_FILTERING

        node = builder.build_datasource_scan_node(
            dialect=dialect,
            compute_mode=compute_mode,
            ds=data_source,
            for_stream=False,
            start_time=start_time,
            end_time=end_time,
        )
        return TectonDataFrame._create(node)
    elif apply_translator:
        timestamp_key = data_source.batch_source.timestamp_field
        if not timestamp_key and (start_time or end_time):
            raise errors.DS_DATAFRAME_NO_TIMESTAMP

        node = builder.build_datasource_scan_node(
            dialect=dialect,
            compute_mode=compute_mode,
            ds=data_source,
            for_stream=False,
            start_time=start_time,
            end_time=end_time,
        )
        return TectonDataFrame._create(node)
    else:
        if start_time is not None or end_time is not None:
            raise errors.DS_RAW_DATAFRAME_NO_TIMESTAMP_FILTER

        node = nodes.RawDataSourceScanNode(dialect, compute_mode, data_source).as_ref()
        return TectonDataFrame._create(node)


def run_transformation_batch(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fd: FeatureDefinitionWrapper,
    feature_start_time: datetime,
    feature_end_time: datetime,
    mock_data_sources: Dict[str, NodeRef],
) -> TectonDataFrame:
    if not fd.is_rtfv_or_prompt:
        check_spark_version(fd.fv_spec.batch_cluster_config)

    feature_time_limits_aligned = pendulum.period(feature_start_time, feature_end_time)

    if fd.is_temporal:
        qt = build_materialization_querytree(
            dialect,
            compute_mode,
            fd,
            for_stream=False,
            feature_data_time_limits=feature_time_limits_aligned,
            use_timestamp_key=True,
        )
    elif fd.is_temporal_aggregate:
        qt = build_pipeline_querytree(
            dialect, compute_mode, fd, for_stream=False, feature_data_time_limits=feature_time_limits_aligned
        )
    else:
        msg = "Unsupported FV Type"
        raise Exception(msg)

    MockDataRewrite(mock_data_sources).rewrite(qt)

    return TectonDataFrame._create(qt)



import multiprocessing

import attrs
import boto3
import pyarrow.parquet as pq

from tecton_core import conf
from tecton_core.compute_mode import ComputeMode
from tecton_core.data_processing_utils import split_spine
from tecton_core.offline_store import DATASET_PARTITION_SIZE
from tecton_core.query.dialect import Dialect
from tecton_core.query.executor_params import QueryTreeStep
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.nodes import PandasDataframeWrapper
from tecton_core.query.nodes import StagingNode
from tecton_core.query.query_tree_executor import QueryTreeExecutor
from tecton_core.query.retrieval_params import GetFeaturesForEventsParams
from tecton_core.query.retrieval_params import GetFeaturesInRangeParams
from tecton_core.query.rewrite import rewrite_tree_for_spine
from tecton_core.query_consts import valid_to
from tecton_materialization.common.dataset_generation import get_features_from_params
from tecton_materialization.common.task_params import get_features_params_from_task_params
from tecton_materialization.ray.delta import OfflineStoreParams
from tecton_materialization.ray.job_status import JobStatusClient
from tecton_materialization.ray.materialization_utils import get_delta_writer
from tecton_materialization.ray.nodes import AddTimePartitionNode
from tecton_materialization.ray.nodes import TimeSpec
from tecton_proto.materialization.job_metadata__client_pb2 import TectonManagedStage
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams


# fewer threads means more memory per thread
DEFAULT_NUM_THREADS = multiprocessing.cpu_count() // 2


def run_dataset_generation(
    materialization_task_params: MaterializationTaskParams,
    job_status_client: JobStatusClient,
    executor: QueryTreeExecutor,
):
    # retrieve current region
    conf.set("CLUSTER_REGION", boto3.Session().region_name)

    assert materialization_task_params.dataset_generation_task_info.HasField("dataset_generation_parameters")
    dataset_generation_params = materialization_task_params.dataset_generation_task_info.dataset_generation_parameters
    params = get_features_params_from_task_params(materialization_task_params, compute_mode=ComputeMode.RIFT)

    num_threads = dataset_generation_params.extra_config.get("num_threads", DEFAULT_NUM_THREADS)
    conf.set("DUCKDB_NTHREADS", str(num_threads))

    qts = []

    if isinstance(params, GetFeaturesForEventsParams):
        time_column = params.timestamp_key
        spine_data = pq.read_table(params.events).to_pandas()
        if conf.get_bool("DUCKDB_ENABLE_SPINE_SPLIT"):
            spine_split = split_spine(spine_data, params.join_keys)
            for spine_chunk in spine_split:
                qt = get_features_from_params(params, spine=PandasDataframeWrapper(spine_chunk))
                qts.append(qt)
        else:
            qts.append(get_features_from_params(params, spine=PandasDataframeWrapper(spine_data)))
    elif isinstance(params, GetFeaturesInRangeParams):
        time_column = valid_to()
        entities = (
            PandasDataframeWrapper(pq.read_table(params.entities).to_pandas()) if params.entities is not None else None
        )
        qts.append(get_features_from_params(params, entities=entities))
    else:
        error = f"Unsupported params type: {type(params)}"
        raise ValueError(error)

    qts = [_add_partition_column(qt, time_column) for qt in qts]

    store_params = OfflineStoreParams(
        feature_view_id=params.fco.id,
        feature_view_name=params.fco.name,
        schema=dataset_generation_params.expected_schema,
        time_spec=None,
        feature_store_format_version=None,
        batch_schedule=None,
    )

    table = get_delta_writer(
        materialization_task_params,
        store_params=store_params,
        table_uri=dataset_generation_params.result_path,
        join_keys=params.join_keys,
    )
    offline_stage_monitor = job_status_client.create_stage_monitor(
        TectonManagedStage.StageType.OFFLINE_STORE,
        "Store results to dataset location",
    )
    for idx, qt in enumerate(qts):
        job_status_client.set_query_index(idx, len(qts))

        rewrite_tree_for_spine(qt)
        reader = executor.exec_qt(qt).result_table

        with offline_stage_monitor():
            table.write(reader)

    table.commit()


def _add_partition_column(qt: NodeRef, time_column) -> NodeRef:
    """
    Injects AddTimePartitionNode either before StagingNode(step=AGGREGATION) or at the top of the tree.
    The aim is to run this node before ODFV (if it is present) to make it part of DuckDB query.
    """

    def create_node(input_node: NodeRef) -> NodeRef:
        return AddTimePartitionNode(
            dialect=Dialect.DUCKDB,
            compute_mode=ComputeMode.RIFT,
            input_node=input_node,
            time_spec=TimeSpec(
                timestamp_key=time_column,
                time_column=time_column,
                partition_size=DATASET_PARTITION_SIZE,
                partition_is_anchor=False,
            ),
        ).as_ref()

    def inject(tree: NodeRef) -> bool:
        """
        Traverse over the tree and return True if AddTimePartitionNode was injected before StagingNode(step=AGGREGATION)
        """
        injected = False

        if isinstance(tree.node, StagingNode) and QueryTreeStep.AGGREGATION == tree.node.query_tree_step:
            prev_input = tree.node.input_node
            new_input = create_node(prev_input)
            tree.node = attrs.evolve(tree.node, input_node=new_input)
            injected = True

        return injected or any(inject(tree=i) for i in tree.inputs)

    if inject(qt):
        return qt

    # add node at the top
    return create_node(qt)



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from typing import ClassVar

DESCRIPTOR: _descriptor.FileDescriptor
SERVER_GROUP_TYPE_FEATURE_SERVER_GROUP: ServerGroupType
SERVER_GROUP_TYPE_TRANSFORM_SERVER_GROUP: ServerGroupType
SERVER_GROUP_TYPE_UNSPECIFIED: ServerGroupType

class ServerGroupType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from tecton._internals import metadata_service
from tecton_core.embeddings.model_artifacts import ModelArtifactProvider
from tecton_proto.common.id__client_pb2 import Id
from tecton_proto.modelartifactservice.model_artifact_service__client_pb2 import FetchModelArtifactRequest


class MDSModelArtifactProvider(ModelArtifactProvider):
    def get_presigned_url(self, model_artifact_id: Id, s3_path: str) -> str:
        try:
            mds = metadata_service.instance()
            resp = mds.FetchModelArtifact(FetchModelArtifactRequest(id=model_artifact_id))
            return resp.model_artifact_download_url
        except Exception as e:
            msg = f"Failed to fetch model artifact {model_artifact_id}: {e}"
            raise RuntimeError(msg)



from google.protobuf import duration_pb2 as _duration_pb2
from tecton_proto.common import column_type__client_pb2 as _column_type__client_pb2
from tecton_proto.common import data_type__client_pb2 as _data_type__client_pb2
from tecton_proto.common import time_window__client_pb2 as _time_window__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Union

DESCRIPTOR: _descriptor.FileDescriptor

class Column(_message.Message):
    __slots__ = ["feature_server_data_type", "feature_server_type", "name", "offline_data_type", "raw_snowflake_type", "raw_spark_type"]
    FEATURE_SERVER_DATA_TYPE_FIELD_NUMBER: ClassVar[int]
    FEATURE_SERVER_TYPE_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    OFFLINE_DATA_TYPE_FIELD_NUMBER: ClassVar[int]
    RAW_SNOWFLAKE_TYPE_FIELD_NUMBER: ClassVar[int]
    RAW_SPARK_TYPE_FIELD_NUMBER: ClassVar[int]
    feature_server_data_type: _data_type__client_pb2.DataType
    feature_server_type: _column_type__client_pb2.ColumnType
    name: str
    offline_data_type: _data_type__client_pb2.DataType
    raw_snowflake_type: str
    raw_spark_type: str
    def __init__(self, name: Optional[str] = ..., offline_data_type: Optional[Union[_data_type__client_pb2.DataType, Mapping]] = ..., feature_server_data_type: Optional[Union[_data_type__client_pb2.DataType, Mapping]] = ..., raw_spark_type: Optional[str] = ..., raw_snowflake_type: Optional[str] = ..., feature_server_type: Optional[Union[_column_type__client_pb2.ColumnType, str]] = ...) -> None: ...

class Field(_message.Message):
    __slots__ = ["dtype", "name"]
    DTYPE_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    dtype: _data_type__client_pb2.DataType
    name: str
    def __init__(self, name: Optional[str] = ..., dtype: Optional[Union[_data_type__client_pb2.DataType, Mapping]] = ...) -> None: ...

class OnlineBatchTableFormat(_message.Message):
    __slots__ = ["online_batch_table_parts"]
    ONLINE_BATCH_TABLE_PARTS_FIELD_NUMBER: ClassVar[int]
    online_batch_table_parts: _containers.RepeatedCompositeFieldContainer[OnlineBatchTablePart]
    def __init__(self, online_batch_table_parts: Optional[Iterable[Union[OnlineBatchTablePart, Mapping]]] = ...) -> None: ...

class OnlineBatchTablePart(_message.Message):
    __slots__ = ["schema", "tiles", "time_window", "window_index"]
    SCHEMA_FIELD_NUMBER: ClassVar[int]
    TILES_FIELD_NUMBER: ClassVar[int]
    TIME_WINDOW_FIELD_NUMBER: ClassVar[int]
    WINDOW_INDEX_FIELD_NUMBER: ClassVar[int]
    schema: Schema
    tiles: _containers.RepeatedCompositeFieldContainer[OnlineBatchTablePartTile]
    time_window: _time_window__client_pb2.TimeWindow
    window_index: int
    def __init__(self, window_index: Optional[int] = ..., time_window: Optional[Union[_time_window__client_pb2.TimeWindow, Mapping]] = ..., schema: Optional[Union[Schema, Mapping]] = ..., tiles: Optional[Iterable[Union[OnlineBatchTablePartTile, Mapping]]] = ...) -> None: ...

class OnlineBatchTablePartTile(_message.Message):
    __slots__ = ["relative_end_time_exclusive", "relative_start_time_inclusive"]
    RELATIVE_END_TIME_EXCLUSIVE_FIELD_NUMBER: ClassVar[int]
    RELATIVE_START_TIME_INCLUSIVE_FIELD_NUMBER: ClassVar[int]
    relative_end_time_exclusive: _duration_pb2.Duration
    relative_start_time_inclusive: _duration_pb2.Duration
    def __init__(self, relative_start_time_inclusive: Optional[Union[_duration_pb2.Duration, Mapping]] = ..., relative_end_time_exclusive: Optional[Union[_duration_pb2.Duration, Mapping]] = ...) -> None: ...

class Schema(_message.Message):
    __slots__ = ["columns"]
    COLUMNS_FIELD_NUMBER: ClassVar[int]
    columns: _containers.RepeatedCompositeFieldContainer[Column]
    def __init__(self, columns: Optional[Iterable[Union[Column, Mapping]]] = ...) -> None: ...



import datetime
import functools
import itertools
import re
import typing
from typing import List
from typing import Union

import tecton_core.tecton_pendulum as pendulum
from tecton_core import specs
from tecton_core.compute_mode import ComputeMode
from tecton_core.fco_container import DataProto
from tecton_core.fco_container import create_fco_container
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.feature_set_config import FeatureSetConfig
from tecton_core.query.retrieval_params import GetFeaturesForEventsParams
from tecton_core.query.retrieval_params import GetFeaturesInRangeParams
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams


SENSITIVE_FIELDS = ["authenticationToken", "secret_access_api_key"]


@functools.lru_cache(maxsize=0)
def feature_definition_from_task_params(params: MaterializationTaskParams) -> FeatureDefinitionWrapper:
    if not params.HasField("feature_view"):
        msg = "'feature_view' must be set in MaterializationTaskParams"
        raise ValueError(msg)

    fco_container = create_fco_container(
        itertools.chain(
            *(
                # The cast helps out various type inference implementations which otherwise would complain that
                # the different FCO lists are not the same type
                typing.cast(typing.Iterable[DataProto], p)
                for p in (params.virtual_data_sources, params.transformations, params.entities)
            )
        ),
        deserialize_funcs_to_main=True,
    )
    fv_spec = specs.create_feature_view_spec_from_data_proto(params.feature_view)
    return FeatureDefinitionWrapper(fv_spec, fco_container)


@functools.lru_cache(maxsize=0)
def feature_service_from_task_params(params: MaterializationTaskParams) -> specs.FeatureServiceSpec:
    if not params.HasField("feature_service"):
        msg = "'feature_service' must be set in MaterializationTaskParams"
        raise ValueError(msg)

    return specs.FeatureServiceSpec.from_data_proto(params.feature_service)


@functools.lru_cache(maxsize=0)
def feature_set_config_from_task_params(params: MaterializationTaskParams) -> FeatureSetConfig:
    if not params.HasField("feature_service"):
        msg = "'feature_service' must be set in MaterializationTaskParams"
        raise ValueError(msg)

    fco_container = create_fco_container(
        itertools.chain(
            *(
                # The cast helps out various type inference implementations which otherwise would complain that
                # the different FCO lists are not the same type
                typing.cast(typing.Iterable[DataProto], p)
                for p in (params.virtual_data_sources, params.transformations, params.entities, params.feature_views)
            )
        ),
        deserialize_funcs_to_main=True,
    )
    fs_spec = feature_service_from_task_params(params)
    return FeatureSetConfig.from_feature_service_spec(fs_spec, fco_container)


def get_features_params_from_task_params(
    params: MaterializationTaskParams,
    compute_mode: ComputeMode,
) -> Union[GetFeaturesInRangeParams, GetFeaturesForEventsParams]:
    dataset_generation_params = params.dataset_generation_task_info.dataset_generation_parameters
    if dataset_generation_params.WhichOneof("input") == "spine":
        spine = dataset_generation_params.spine

        if params.HasField("feature_view"):
            fco = feature_definition_from_task_params(params)
            feature_set_config = None
        elif params.HasField("feature_service"):
            fco = feature_service_from_task_params(params)
            feature_set_config = feature_set_config_from_task_params(params)
        else:
            msg = "Either feature_view or feature_service must be set in MaterializationTaskParams"
            raise ValueError(msg)

        return GetFeaturesForEventsParams(
            fco=fco,
            events=spine.path,
            timestamp_key=spine.timestamp_key if spine.HasField("timestamp_key") else None,
            from_source=dataset_generation_params.from_source
            if dataset_generation_params.HasField("from_source") and dataset_generation_params.from_source == True
            else None,
            compute_mode=compute_mode,
            feature_set_config=feature_set_config,
        )
    elif dataset_generation_params.WhichOneof("input") == "datetime_range":
        datetime_range = dataset_generation_params.datetime_range
        start_time = datetime_range.start.ToDatetime()
        end_time = datetime_range.end.ToDatetime()
        max_lookback_time = (
            datetime_range.max_lookback.ToDatetime() if datetime_range.HasField("max_lookback") else None
        )
        entities = datetime_range.entities_path if datetime_range.HasField("entities_path") else None
        return GetFeaturesInRangeParams(
            fco=feature_definition_from_task_params(params),
            start_time=start_time,
            end_time=end_time,
            entities=entities,
            max_lookback=(max_lookback_time - start_time) if max_lookback_time else None,
            from_source=dataset_generation_params.from_source
            if dataset_generation_params.HasField("from_source")
            else None,
            compute_mode=compute_mode,
        )
    else:
        error = "Invalid dataset_generation_params"
        raise ValueError(error)


class TimeInterval(typing.NamedTuple):
    start: datetime.datetime
    end: datetime.datetime

    def to_pendulum(self) -> pendulum.Period:
        return pendulum.instance(self.end) - pendulum.instance(self.start)


def _backfill_job_periods(
    start_time: datetime.datetime, end_time: datetime.datetime, interval: datetime.timedelta
) -> List[TimeInterval]:
    jobs = []
    while start_time < end_time:
        jobs.append(TimeInterval(start_time, start_time + interval))
        start_time = start_time + interval
    assert start_time == end_time, "Start and end times were not aligned to `interval`"
    return jobs


def job_query_intervals(task_params: MaterializationTaskParams) -> typing.List[TimeInterval]:
    """
    Return a list of start/end tuples of size batch_schedule.
    For use of breaking up a large backfill window into incremental sizes.
    """
    fd = feature_definition_from_task_params(task_params)
    assert task_params.batch_task_info.HasField("batch_parameters")
    task_feature_start_time = task_params.batch_task_info.batch_parameters.feature_start_time.ToDatetime()
    task_feature_end_time = task_params.batch_task_info.batch_parameters.feature_end_time.ToDatetime()
    # for incremental backfills, we split the job into each batch interval
    if fd.is_incremental_backfill:
        return _backfill_job_periods(task_feature_start_time, task_feature_end_time, fd.batch_materialization_schedule)
    else:
        return [TimeInterval(task_feature_start_time, task_feature_end_time)]


def redact_sensitive_fields_from_params(message_json: str):
    for field_name in SENSITIVE_FIELDS:
        message_json = re.sub(rf'"{field_name}": ".*?"', f'"{field_name}": "<REDACTED>"', message_json)
    return message_json



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from tecton_proto.common import container_image__client_pb2 as _container_image__client_pb2
from tecton_proto.common import id__client_pb2 as _id__client_pb2
from tecton_proto.common import server_group_status__client_pb2 as _server_group_status__client_pb2
from tecton_proto.common import server_group_type__client_pb2 as _server_group_type__client_pb2
from tecton_proto.realtime import instance_group__client_pb2 as _instance_group__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Mapping, Optional, Union

DESCRIPTOR: _descriptor.FileDescriptor

class AutoscalingPolicy(_message.Message):
    __slots__ = ["autoscaling_enabled", "initialization_period_seconds", "target_concurrent_request_limit", "target_cpu_utilization", "target_percentage_node_utilization"]
    AUTOSCALING_ENABLED_FIELD_NUMBER: ClassVar[int]
    INITIALIZATION_PERIOD_SECONDS_FIELD_NUMBER: ClassVar[int]
    TARGET_CONCURRENT_REQUEST_LIMIT_FIELD_NUMBER: ClassVar[int]
    TARGET_CPU_UTILIZATION_FIELD_NUMBER: ClassVar[int]
    TARGET_PERCENTAGE_NODE_UTILIZATION_FIELD_NUMBER: ClassVar[int]
    autoscaling_enabled: bool
    initialization_period_seconds: int
    target_concurrent_request_limit: int
    target_cpu_utilization: float
    target_percentage_node_utilization: int
    def __init__(self, autoscaling_enabled: bool = ..., target_cpu_utilization: Optional[float] = ..., target_concurrent_request_limit: Optional[int] = ..., target_percentage_node_utilization: Optional[int] = ..., initialization_period_seconds: Optional[int] = ...) -> None: ...

class FeatureServerGroupState(_message.Message):
    __slots__ = ["instance_type"]
    INSTANCE_TYPE_FIELD_NUMBER: ClassVar[int]
    instance_type: str
    def __init__(self, instance_type: Optional[str] = ...) -> None: ...

class ServerGroupState(_message.Message):
    __slots__ = ["autoscaling_policy", "created_at", "desired_nodes", "feature_server_group_state", "last_updated_at", "max_nodes", "min_nodes", "name", "server_group_id", "server_group_state_id", "status", "status_details", "transform_server_group_state", "type", "workspace", "workspace_state_id"]
    AUTOSCALING_POLICY_FIELD_NUMBER: ClassVar[int]
    CREATED_AT_FIELD_NUMBER: ClassVar[int]
    DESIRED_NODES_FIELD_NUMBER: ClassVar[int]
    FEATURE_SERVER_GROUP_STATE_FIELD_NUMBER: ClassVar[int]
    LAST_UPDATED_AT_FIELD_NUMBER: ClassVar[int]
    MAX_NODES_FIELD_NUMBER: ClassVar[int]
    MIN_NODES_FIELD_NUMBER: ClassVar[int]
    NAME_FIELD_NUMBER: ClassVar[int]
    SERVER_GROUP_ID_FIELD_NUMBER: ClassVar[int]
    SERVER_GROUP_STATE_ID_FIELD_NUMBER: ClassVar[int]
    STATUS_DETAILS_FIELD_NUMBER: ClassVar[int]
    STATUS_FIELD_NUMBER: ClassVar[int]
    TRANSFORM_SERVER_GROUP_STATE_FIELD_NUMBER: ClassVar[int]
    TYPE_FIELD_NUMBER: ClassVar[int]
    WORKSPACE_FIELD_NUMBER: ClassVar[int]
    WORKSPACE_STATE_ID_FIELD_NUMBER: ClassVar[int]
    autoscaling_policy: AutoscalingPolicy
    created_at: _timestamp_pb2.Timestamp
    desired_nodes: int
    feature_server_group_state: FeatureServerGroupState
    last_updated_at: _timestamp_pb2.Timestamp
    max_nodes: int
    min_nodes: int
    name: str
    server_group_id: _id__client_pb2.Id
    server_group_state_id: _id__client_pb2.Id
    status: _server_group_status__client_pb2.ServerGroupStatus
    status_details: str
    transform_server_group_state: TransformServerGroupState
    type: _server_group_type__client_pb2.ServerGroupType
    workspace: str
    workspace_state_id: _id__client_pb2.Id
    def __init__(self, server_group_state_id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., server_group_id: Optional[Union[_id__client_pb2.Id, Mapping]] = ..., name: Optional[str] = ..., status: Optional[Union[_server_group_status__client_pb2.ServerGroupStatus, str]] = ..., type: Optional[Union[_server_group_type__client_pb2.ServerGroupType, str]] = ..., autoscaling_policy: Optional[Union[AutoscalingPolicy, Mapping]] = ..., status_details: Optional[str] = ..., created_at: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ..., last_updated_at: Optional[Union[_timestamp_pb2.Timestamp, Mapping]] = ..., min_nodes: Optional[int] = ..., max_nodes: Optional[int] = ..., desired_nodes: Optional[int] = ..., transform_server_group_state: Optional[Union[TransformServerGroupState, Mapping]] = ..., feature_server_group_state: Optional[Union[FeatureServerGroupState, Mapping]] = ..., workspace: Optional[str] = ..., workspace_state_id: Optional[Union[_id__client_pb2.Id, Mapping]] = ...) -> None: ...

class TransformServerGroupState(_message.Message):
    __slots__ = ["autoscaling_policy", "environment_id", "environment_name", "environment_variables", "image_info", "instance_group_handle"]
    class EnvironmentVariablesEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: str
        def __init__(self, key: Optional[str] = ..., value: Optional[str] = ...) -> None: ...
    AUTOSCALING_POLICY_FIELD_NUMBER: ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: ClassVar[int]
    ENVIRONMENT_NAME_FIELD_NUMBER: ClassVar[int]
    ENVIRONMENT_VARIABLES_FIELD_NUMBER: ClassVar[int]
    IMAGE_INFO_FIELD_NUMBER: ClassVar[int]
    INSTANCE_GROUP_HANDLE_FIELD_NUMBER: ClassVar[int]
    autoscaling_policy: AutoscalingPolicy
    environment_id: str
    environment_name: str
    environment_variables: _containers.ScalarMap[str, str]
    image_info: _container_image__client_pb2.ContainerImage
    instance_group_handle: _instance_group__client_pb2.InstanceGroupHandle
    def __init__(self, environment_id: Optional[str] = ..., environment_name: Optional[str] = ..., image_info: Optional[Union[_container_image__client_pb2.ContainerImage, Mapping]] = ..., environment_variables: Optional[Mapping[str, str]] = ..., autoscaling_policy: Optional[Union[AutoscalingPolicy, Mapping]] = ..., instance_group_handle: Optional[Union[_instance_group__client_pb2.InstanceGroupHandle, Mapping]] = ...) -> None: ...



from google.protobuf import descriptor_pb2 as _descriptor_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Union

AUTH_METADATA_FIELD_NUMBER: ClassVar[int]
DESCRIPTOR: _descriptor.FileDescriptor
RESOURCE_REF_TYPE_PRINCIPAL_GROUP_ID: ResourceRefTypeEnum
RESOURCE_REF_TYPE_SECRET_SCOPE: ResourceRefTypeEnum
RESOURCE_REF_TYPE_SERVICE_ACCOUNT_ID: ResourceRefTypeEnum
RESOURCE_REF_TYPE_UNSPECIFIED: ResourceRefTypeEnum
RESOURCE_REF_TYPE_WORKSPACE_NAME: ResourceRefTypeEnum
auth_metadata: _descriptor.FieldDescriptor

class AuthMetadata(_message.Message):
    __slots__ = ["advanced_permission_overrides", "defer_authorization_to_service", "permission", "resource_reference", "skip_authentication", "skip_authorization"]
    ADVANCED_PERMISSION_OVERRIDES_FIELD_NUMBER: ClassVar[int]
    DEFER_AUTHORIZATION_TO_SERVICE_FIELD_NUMBER: ClassVar[int]
    PERMISSION_FIELD_NUMBER: ClassVar[int]
    RESOURCE_REFERENCE_FIELD_NUMBER: ClassVar[int]
    SKIP_AUTHENTICATION_FIELD_NUMBER: ClassVar[int]
    SKIP_AUTHORIZATION_FIELD_NUMBER: ClassVar[int]
    advanced_permission_overrides: _containers.RepeatedCompositeFieldContainer[PermissionOverride]
    defer_authorization_to_service: bool
    permission: str
    resource_reference: ResourceReference
    skip_authentication: bool
    skip_authorization: bool
    def __init__(self, skip_authentication: bool = ..., skip_authorization: bool = ..., permission: Optional[str] = ..., advanced_permission_overrides: Optional[Iterable[Union[PermissionOverride, Mapping]]] = ..., resource_reference: Optional[Union[ResourceReference, Mapping]] = ..., defer_authorization_to_service: bool = ...) -> None: ...

class PermissionOverride(_message.Message):
    __slots__ = ["condition_field_path", "condition_value", "permission_override"]
    CONDITION_FIELD_PATH_FIELD_NUMBER: ClassVar[int]
    CONDITION_VALUE_FIELD_NUMBER: ClassVar[int]
    PERMISSION_OVERRIDE_FIELD_NUMBER: ClassVar[int]
    condition_field_path: str
    condition_value: str
    permission_override: str
    def __init__(self, condition_field_path: Optional[str] = ..., condition_value: Optional[str] = ..., permission_override: Optional[str] = ...) -> None: ...

class ResourceReference(_message.Message):
    __slots__ = ["path", "type"]
    PATH_FIELD_NUMBER: ClassVar[int]
    TYPE_FIELD_NUMBER: ClassVar[int]
    path: str
    type: ResourceRefTypeEnum
    def __init__(self, type: Optional[Union[ResourceRefTypeEnum, str]] = ..., path: Optional[str] = ...) -> None: ...

class ResourceRefTypeEnum(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from google.protobuf import duration_pb2 as _duration_pb2
from tecton_proto.dataobs import expectation__client_pb2 as _expectation__client_pb2
from tecton_proto.dataobs import metric__client_pb2 as _metric__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Union

DESCRIPTOR: _descriptor.FileDescriptor

class DataObservabilityConfig(_message.Message):
    __slots__ = ["feature_expectation_validation_schedule", "feature_expectations", "feature_view_name", "metric_expectations", "metrics", "workspace"]
    FEATURE_EXPECTATIONS_FIELD_NUMBER: ClassVar[int]
    FEATURE_EXPECTATION_VALIDATION_SCHEDULE_FIELD_NUMBER: ClassVar[int]
    FEATURE_VIEW_NAME_FIELD_NUMBER: ClassVar[int]
    METRICS_FIELD_NUMBER: ClassVar[int]
    METRIC_EXPECTATIONS_FIELD_NUMBER: ClassVar[int]
    WORKSPACE_FIELD_NUMBER: ClassVar[int]
    feature_expectation_validation_schedule: str
    feature_expectations: _containers.RepeatedCompositeFieldContainer[_expectation__client_pb2.FeatureExpectation]
    feature_view_name: str
    metric_expectations: _containers.RepeatedCompositeFieldContainer[_expectation__client_pb2.MetricExpectation]
    metrics: _containers.RepeatedCompositeFieldContainer[_metric__client_pb2.Metric]
    workspace: str
    def __init__(self, workspace: Optional[str] = ..., feature_view_name: Optional[str] = ..., feature_expectation_validation_schedule: Optional[str] = ..., metrics: Optional[Iterable[Union[_metric__client_pb2.Metric, Mapping]]] = ..., feature_expectations: Optional[Iterable[Union[_expectation__client_pb2.FeatureExpectation, Mapping]]] = ..., metric_expectations: Optional[Iterable[Union[_expectation__client_pb2.MetricExpectation, Mapping]]] = ...) -> None: ...

class DataObservabilityMaterializationConfig(_message.Message):
    __slots__ = ["enabled", "metric_interval", "metric_table_name"]
    ENABLED_FIELD_NUMBER: ClassVar[int]
    METRIC_INTERVAL_FIELD_NUMBER: ClassVar[int]
    METRIC_TABLE_NAME_FIELD_NUMBER: ClassVar[int]
    enabled: bool
    metric_interval: _duration_pb2.Duration
    metric_table_name: str
    def __init__(self, enabled: bool = ..., metric_interval: Optional[Union[_duration_pb2.Duration, Mapping]] = ..., metric_table_name: Optional[str] = ...) -> None: ...



from datetime import datetime
from typing import Optional

import pandas


REQUEST_TIMESTAMP_FIELD_NAME = "request_timestamp"


class RealtimeContext:
    """
    RealtimeContext is a class that is used to pass context metadata such as the request_timestamp
    to the `context` parameter of a Realtime Feature Views.
    """

    _is_python_mode: Optional[bool] = None
    _request_timestamp: Optional[datetime] = None
    _row_level_data: Optional[pandas.DataFrame] = None

    def __init__(
        self,
        request_timestamp: Optional[datetime] = None,
        row_level_data: Optional[pandas.DataFrame] = None,
        _is_python_mode: Optional[bool] = None,
    ) -> None:
        """
        Initialize the RealtimeContext object.

        :param _is_python_mode: Whether the Realtime Feature View is in Python mode.
        :param request_timestamp: The timestamp of the request made to the Tecton Feature Server. Used for Python mode offline and for both modes online.
        :param row_level_data: The row-level context data for the Realtime Feature View for each row in the events data frame. Only populated in Pandas mode during Offline Retrieval.
        """
        self._is_python_mode = _is_python_mode
        self._request_timestamp = request_timestamp
        self._row_level_data = row_level_data

    def set_mode(self, is_python: bool) -> None:
        """
        Set the mode of the Realtime Feature View.

        :param is_python: Whether the Realtime Feature View is in Python mode.
        """
        self._is_python_mode = is_python

    @property
    def request_timestamp(self) -> Optional[datetime]:
        """
        The request_timestamp is the timestamp of the request made to the Tecton Feature Server.
        """
        if not self._is_python_mode:
            str = "The field `request_timestamp` is only available in Realtime Feature Views with Python mode. Please use `context.request_timestamp_series` when using Pandas mode."
            raise ValueError(str)

        return self._request_timestamp

    @property
    def request_timestamp_series(self) -> Optional[pandas.Series]:
        """
        The request_timestamp_series is a pandas Series with the request_timestamp as the only element. Each element in the
        Series is the request_timestamp for the corresponding row in the input data.
        """
        if self._is_python_mode:
            str = "The field `request_timestamp_series` is only available in Realtime Feature Views with Pandas mode. Please use `context.request_timestamp` when using Python mode."
            raise ValueError(str)

        # For online retrieval, we build a Series with the request_timestamp as the only element
        if self._request_timestamp is not None:
            return pandas.Series([self._request_timestamp], name=REQUEST_TIMESTAMP_FIELD_NAME)

        if self._row_level_data is None:
            return None

        return self._row_level_data[REQUEST_TIMESTAMP_FIELD_NAME]



import copy
import functools
import sys
import typing as t
from dataclasses import dataclass
from typing import Callable
from typing import List
from typing import Optional

import click
from click import Command
from click import Context
from click import HelpFormatter
from click import echo
from click import style

import tecton_core.tecton_pendulum as pendulum
from tecton import tecton_context
from tecton._internals.analytics import AnalyticsLogger
from tecton._internals.analytics import StateUpdateEventMetrics
from tecton._internals.analytics import StateUpdateResult
from tecton._internals.utils import cluster_url
from tecton.cli import printer
from tecton.cli import workspace_utils
from tecton_core import conf


analytics = AnalyticsLogger()


class TectonCommand(click.Command):
    # used by integration tests
    skip_config_check = False

    """Base class for Click which implements required_auth and uses_workspace behavior, as well as analytics."""

    def __init__(
        self,
        *args,
        callback,
        requires_auth=True,
        uses_workspace=False,
        is_main_command=False,
        deprecation_warning="",
        **kwargs,
    ):
        self.is_main_command = is_main_command
        self.deprecation_warning = deprecation_warning

        @functools.wraps(callback)
        @click.pass_context
        def wrapper(ctx, *cbargs, **cbkwargs):
            host = cluster_url()
            cluster_configured = host is not None

            command_names = []
            cur = ctx
            # The top level context is `cli` which we don't want to include.
            while cur:
                command_names.append(cur.command.name)
                cur = cur.parent
            command_names.reverse()

            # TODO(TEC-14547): Move `workspace` param usages to callback function.
            is_create_workspace_command = ctx.parent.info_name == "workspace" and ctx.info_name == "create"
            has_workspace_option = ctx.params.get("workspace", None) is not None
            if has_workspace_option and not is_create_workspace_command:
                workspace_name = ctx.params["workspace"]
                workspace_utils.check_workspace_exists(workspace_name)
                conf.set("TECTON_WORKSPACE", workspace_name)

            # Do not try logging events if cluster has never be configured or if user is trying to log in,
            # otherwise the CLI either won't be able to find the MDS or auth token might have expired
            if cluster_configured:
                if uses_workspace:
                    printer.safe_print(f'Using workspace "{tecton_context.get_current_workspace()}" on cluster {host}')
                start_time = pendulum.now("UTC")
                state_update_event = None
                try:
                    invoke_result = ctx.invoke(callback, *cbargs, **cbkwargs)
                    if isinstance(invoke_result, StateUpdateResult):
                        state_update_event = invoke_result.state_update_event_metrics
                except PermissionError as e:
                    state_update_event = StateUpdateEventMetrics.from_error_message(str(e))
                    printer.safe_print(f"Unable to execute command `{' '.join(command_names)}`: {str(e)}")
                execution_time = pendulum.now("UTC") - start_time
                if requires_auth:
                    if state_update_event:
                        # TODO: Include sub-command?
                        analytics.log_cli_event(command_names[1], execution_time, ctx.params, state_update_event)
                        if state_update_event.error_message:
                            sys.exit(1)
                    else:
                        analytics.log_cli_event(command_names[1], execution_time, ctx.params)
            elif not requires_auth or TectonCommand.skip_config_check:
                # Do not try executing anything besides unauthenticated commnds (`login`, `version`) when cluster hasn't been configured.
                state_update_event = ctx.invoke(callback, *cbargs, **cbkwargs)
            else:
                printer.safe_print(
                    f"`{' '.join(command_names)}` requires authentication. Please authenticate using `tecton login`."
                )
                sys.exit(1)

        super().__init__(*args, callback=wrapper, **kwargs)

    def invoke(self, ctx: Context) -> t.Any:
        """Given a context, this invokes the attached callback (if it exists)
        in the right way.

        Over-writes the base method for a better depredation warning
        """
        if self.deprecated:
            if self.deprecation_warning:
                message = "DeprecationWarning:" + self.deprecation_warning
            else:
                message = f"DeprecationWarning: The command {self.name} is deprecated."
            echo(style(message, fg="red"), err=True)

        if self.callback is not None:
            return ctx.invoke(self.callback, **ctx.params)


class TectonGroup(click.Group):
    """Routes group.command calls to use TectonCommand instead of the base Click command"""

    def __init__(self, is_main_command=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_main_command = is_main_command

    command_class = TectonCommand

    def add_deprecated_command(
        self, cmd: Command, name: t.Optional[str] = None, new_target: t.Optional[str] = None
    ) -> None:
        deprecation_warning = f" The command `{self.name} {name}` is deprecated and will removed in a future version."
        if new_target:
            deprecation_warning += f" Use `{new_target}` instead."
        cmd = copy.copy(cmd)
        cmd.hidden = True
        cmd.deprecated = True
        cmd.deprecation_warning = deprecation_warning
        cmd.help = deprecation_warning + "\n\n" + cmd.help
        super().add_command(cmd, name)


TectonGroup.group_class = TectonGroup


class CategorizedTectonGroup(TectonGroup):
    """A command group that splits main commands into a separate section header from the non-main commnads"""

    def format_commands(self, ctx: Context, formatter: HelpFormatter) -> None:
        main = []
        other = []
        all_commands = self.list_commands(ctx)
        limit = formatter.width - 6 - max(len(name) for name in all_commands)

        for name in all_commands:
            command = self.get_command(ctx, name)

            if command is None or command.hidden:
                continue

            help_record = (name, command.get_short_help_str(limit=limit))
            if command.is_main_command:
                main.append(help_record)
            else:
                other.append(help_record)

        if main:
            with formatter.section("Main commands"):
                formatter.write_dl(main)

        if other:
            with formatter.section("Other commands"):
                formatter.write_dl(other)


@dataclass
class HiddenValue:
    """Wraps a string and displays it as ****. Used to hide passwords when showing the previous value in prompts."""

    value: str

    def __str__(self):
        return "*" * len(self.value)


class HiddenValueType(click.ParamType):
    name = "TEXT"  # Controls how this is displayed in help text

    def convert(self, value, param=None, ctx=None):
        if value is None:
            return None
        if isinstance(value, HiddenValue):
            return value
        return HiddenValue(value)


def tecton_config_option(
    key: str, param_decls: Optional[List[str]] = None, hide_input: bool = False, **kwargs
) -> Callable:
    """Decorator for a Click option which defaults to a value from tecton_core.conf and prompts if the key is not set.

    :arg key The tecton_core.conf key.
    :arg param_decls The Click param_decls.
    :arg hide_input Whether to hide user input in the prompt, as for passwords. Additionally, if a default is set,
         it will be obfuscated with `*` in the prompt.
    """
    if "prompt" not in kwargs:
        kwargs["prompt"] = True
    lower_key = key.lower().replace("_", "-")
    param_decls = param_decls or [f"--{lower_key}"]

    if hide_input:
        kwargs["type"] = HiddenValueType()
        default = HiddenValueType().convert(conf.get_or_none(key))
    else:

        def default():
            return conf.get_or_none(key)

    def decorator(f):
        @click.option(*param_decls, **kwargs, hide_input=hide_input, default=default)
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            def unwrap(v):
                return v.value if isinstance(v, HiddenValue) else v

            return f(**{k: unwrap(v) for k, v in kwargs.items()})

        return wrapper

    return decorator



from typing import Dict
from typing import List

from tecton import DataSource
from tecton import Entity
from tecton import FeatureService
from tecton import FeatureTable
from tecton import FeatureView
from tecton import Prompt
from tecton import Transformation
from tecton.framework.base_tecton_object import BaseTectonObject


class TestRepo:
    """
    A repository class for accessing and managing various Tecton objects.

    This class stores and provides access to FeatureViews, Entities, Data Sources,
    FeatureServices, and Transformations defined in the Feature Repository.
    """

    _feature_views: Dict[str, FeatureView] = {}
    _entities: Dict[str, Entity] = {}
    _data_sources: Dict[str, DataSource] = {}
    _feature_services: Dict[str, FeatureService] = {}
    _transformations: Dict[str, Transformation] = {}

    _type_to_collection: Dict[type, Dict[str, BaseTectonObject]] = {
        FeatureView: _feature_views,
        Entity: _entities,
        DataSource: _data_sources,
        FeatureService: _feature_services,
        Transformation: _transformations,
    }

    def __init__(self, objects: List[BaseTectonObject]):
        """
        Initialize the TestRepo with a list of Tecton objects.

        Args:
            objects (List[BaseTectonObject]): A list of Tecton objects to be stored in the repository.
        """
        for obj in objects:
            for fcoType, collection in self._type_to_collection.items():
                if isinstance(obj, fcoType):
                    collection[obj.name] = obj
                    break

    def _get_fco(self, name: str, fcoType: type) -> BaseTectonObject:
        if name in self._type_to_collection[fcoType]:
            return self._type_to_collection[fcoType][name]
        else:
            err = f"{fcoType.__name__} '{name}' not found."
            raise KeyError(err)

    def get_all_objects(self) -> List[BaseTectonObject]:
        """
        Retrieve all Tecton objects stored in the repository.

        Returns:
            List[BaseTectonObject]: A list of all stored Tecton objects.
        """
        objects = [
            tecton_obj
            for collection_values in self._type_to_collection.values()
            for name, tecton_obj in collection_values.items()
        ]
        return objects

    def get_feature_view(self, name) -> FeatureView:
        """
        Retrieve a FeatureView by name.

        Args:
            name (str): The name of the FeatureView to retrieve.

        Returns:
            FeatureView: The requested FeatureView.

        Raises:
            KeyError: If the FeatureView is not found in the repository.
        """
        return self._get_fco(name, FeatureView)

    def get_feature_table(self, name) -> FeatureTable:
        """
        Retrieve a FeatureTable by name.

        Args:
            name (str): The name of the FeatureTable to retrieve.

        Returns:
            FeatureTable: The requested FeatureTable.

        Raises:
            KeyError: If the FeatureTable is not found in the repository.
        """
        return self._get_fco(name, FeatureView)

    def get_feature_service(self, name) -> FeatureService:
        """
        Retrieve a FeatureService by name.

        Args:
            name (str): The name of the FeatureService to retrieve.

        Returns:
            FeatureService: The requested FeatureService.

        Raises:
            KeyError: If the FeatureService is not found in the repository.
        """
        return self._get_fco(name, FeatureService)

    def get_data_source(self, name) -> DataSource:
        """
        Retrieve a DataSource by name.

        Args:
            name (str): The name of the DataSource to retrieve.

        Returns:
            DataSource: The requested DataSource.

        Raises:
            KeyError: If the DataSource is not found in the repository.
        """
        return self._get_fco(name, DataSource)

    def get_transformation(self, name) -> Transformation:
        """
        Retrieve a Transformation by name.

        Args:
            name (str): The name of the Transformation to retrieve.

        Returns:
            Transformation: The requested Transformation.

        Raises:
            KeyError: If the Transformation is not found in the repository.
        """
        return self._get_fco(name, Transformation)

    def get_entity(self, name: str) -> Entity:
        """
        Retrieve an entity by name.

        :param name: The name of the Transformation to retrieve.

        :return: The requested Entity

        :raises KeyError: If the Entity is not found in the repository.
        """
        return self._get_fco(name, Entity)

    def get_prompt(self, name: str) -> Prompt:
        """
        Retrieve an entity by name.

        :param name: The name of the Transformation to retrieve.

        :return: The requested Entity

        :raises KeyError: If the Entity is not found in the repository.
        """
        return self._get_fco(name, FeatureView)

    @property
    def feature_views(self):
        """Get all FeatureViews in the repository."""
        return [fv for fv in self._feature_views.values() if not isinstance(fv, (FeatureTable, Prompt))]

    @property
    def feature_tables(self):
        """Get all FeatureTables in the repository."""
        fts = [fv for fv in self._feature_views.values() if isinstance(fv, FeatureTable)]
        return fts

    @property
    def prompts(self):
        """Get all Prompts in the repository."""
        fts = [fv for fv in self._feature_views.values() if isinstance(fv, Prompt)]
        return fts

    @property
    def entities(self):
        """Get all Entities in the repository."""
        return list(self._entities.values())

    @property
    def transformations(self):
        """Get all Transformations in the repository."""
        return list(self._transformations.values())

    @property
    def feature_services(self):
        """Get all Feature Services in the repository."""
        return list(self._feature_services.values())

    @property
    def data_sources(self):
        """Get all Data Sources in the repository."""
        return list(self._data_sources.values())



from tecton_proto.args import data_source_config__client_pb2 as _data_source_config__client_pb2
from tecton_proto.args import feature_view__client_pb2 as _feature_view__client_pb2
from tecton_proto.args import user_defined_function__client_pb2 as _user_defined_function__client_pb2
from tecton_proto.args import virtual_data_source__client_pb2 as _virtual_data_source__client_pb2
from tecton_proto.common import schema__client_pb2 as _schema__client_pb2
from tecton_proto.common import spark_schema__client_pb2 as _spark_schema__client_pb2
from tecton_proto.data import feature_view__client_pb2 as _feature_view__client_pb2
from tecton_proto.data import hive_metastore__client_pb2 as _hive_metastore__client_pb2
from tecton_proto.data import transformation__client_pb2 as _transformation__client_pb2
from tecton_proto.data import virtual_data_source__client_pb2 as _virtual_data_source__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar, Iterable, Mapping, Optional, Union

DESCRIPTOR: _descriptor.FileDescriptor

class ExecuteRequest(_message.Message):
    __slots__ = ["envVars", "getBatchDataSourceFunctionSchema", "getFeatureViewSchema", "getFileSourceSchema", "getHiveTableSchema", "getKafkaSourceSchema", "getKinesisSourceSchema", "getMultipleDataSourceSchemasRequest", "getMultipleFeatureViewSchemasRequest", "getQueryPlanInfoForFeatureViewPipeline", "getRedshiftTableSchema", "getSnowflakeSchema", "getStreamDataSourceFunctionSchema", "getUnityTableSchema", "listHiveDatabases", "listHiveTableColumns", "listHiveTables"]
    class EnvVarsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: str
        def __init__(self, key: Optional[str] = ..., value: Optional[str] = ...) -> None: ...
    ENVVARS_FIELD_NUMBER: ClassVar[int]
    GETBATCHDATASOURCEFUNCTIONSCHEMA_FIELD_NUMBER: ClassVar[int]
    GETFEATUREVIEWSCHEMA_FIELD_NUMBER: ClassVar[int]
    GETFILESOURCESCHEMA_FIELD_NUMBER: ClassVar[int]
    GETHIVETABLESCHEMA_FIELD_NUMBER: ClassVar[int]
    GETKAFKASOURCESCHEMA_FIELD_NUMBER: ClassVar[int]
    GETKINESISSOURCESCHEMA_FIELD_NUMBER: ClassVar[int]
    GETMULTIPLEDATASOURCESCHEMASREQUEST_FIELD_NUMBER: ClassVar[int]
    GETMULTIPLEFEATUREVIEWSCHEMASREQUEST_FIELD_NUMBER: ClassVar[int]
    GETQUERYPLANINFOFORFEATUREVIEWPIPELINE_FIELD_NUMBER: ClassVar[int]
    GETREDSHIFTTABLESCHEMA_FIELD_NUMBER: ClassVar[int]
    GETSNOWFLAKESCHEMA_FIELD_NUMBER: ClassVar[int]
    GETSTREAMDATASOURCEFUNCTIONSCHEMA_FIELD_NUMBER: ClassVar[int]
    GETUNITYTABLESCHEMA_FIELD_NUMBER: ClassVar[int]
    LISTHIVEDATABASES_FIELD_NUMBER: ClassVar[int]
    LISTHIVETABLECOLUMNS_FIELD_NUMBER: ClassVar[int]
    LISTHIVETABLES_FIELD_NUMBER: ClassVar[int]
    envVars: _containers.ScalarMap[str, str]
    getBatchDataSourceFunctionSchema: GetBatchDataSourceFunctionSchema
    getFeatureViewSchema: GetFeatureViewSchema
    getFileSourceSchema: GetFileSourceSchema
    getHiveTableSchema: GetHiveTableSchema
    getKafkaSourceSchema: GetKafkaSourceSchema
    getKinesisSourceSchema: GetKinesisSourceSchema
    getMultipleDataSourceSchemasRequest: GetMultipleDataSourceSchemasRequest
    getMultipleFeatureViewSchemasRequest: GetMultipleFeatureViewSchemasRequest
    getQueryPlanInfoForFeatureViewPipeline: GetQueryPlanInfoForFeatureViewPipeline
    getRedshiftTableSchema: GetRedshiftTableSchema
    getSnowflakeSchema: GetSnowflakeSchema
    getStreamDataSourceFunctionSchema: GetStreamDataSourceFunctionSchema
    getUnityTableSchema: GetUnityTableSchema
    listHiveDatabases: ListHiveDatabases
    listHiveTableColumns: ListHiveTableColumns
    listHiveTables: ListHiveTables
    def __init__(self, getMultipleFeatureViewSchemasRequest: Optional[Union[GetMultipleFeatureViewSchemasRequest, Mapping]] = ..., getMultipleDataSourceSchemasRequest: Optional[Union[GetMultipleDataSourceSchemasRequest, Mapping]] = ..., getHiveTableSchema: Optional[Union[GetHiveTableSchema, Mapping]] = ..., getRedshiftTableSchema: Optional[Union[GetRedshiftTableSchema, Mapping]] = ..., getFileSourceSchema: Optional[Union[GetFileSourceSchema, Mapping]] = ..., getKinesisSourceSchema: Optional[Union[GetKinesisSourceSchema, Mapping]] = ..., getKafkaSourceSchema: Optional[Union[GetKafkaSourceSchema, Mapping]] = ..., getFeatureViewSchema: Optional[Union[GetFeatureViewSchema, Mapping]] = ..., getSnowflakeSchema: Optional[Union[GetSnowflakeSchema, Mapping]] = ..., getBatchDataSourceFunctionSchema: Optional[Union[GetBatchDataSourceFunctionSchema, Mapping]] = ..., getStreamDataSourceFunctionSchema: Optional[Union[GetStreamDataSourceFunctionSchema, Mapping]] = ..., getUnityTableSchema: Optional[Union[GetUnityTableSchema, Mapping]] = ..., listHiveDatabases: Optional[Union[ListHiveDatabases, Mapping]] = ..., listHiveTables: Optional[Union[ListHiveTables, Mapping]] = ..., listHiveTableColumns: Optional[Union[ListHiveTableColumns, Mapping]] = ..., getQueryPlanInfoForFeatureViewPipeline: Optional[Union[GetQueryPlanInfoForFeatureViewPipeline, Mapping]] = ..., envVars: Optional[Mapping[str, str]] = ...) -> None: ...

class ExecuteResult(_message.Message):
    __slots__ = ["listHiveResult", "multipleDataSourceSchemaResponse", "multipleFeatureViewSchemaResponse", "queryPlanInfo", "schema", "sparkSchema", "uncaughtError", "validationError"]
    LISTHIVERESULT_FIELD_NUMBER: ClassVar[int]
    MULTIPLEDATASOURCESCHEMARESPONSE_FIELD_NUMBER: ClassVar[int]
    MULTIPLEFEATUREVIEWSCHEMARESPONSE_FIELD_NUMBER: ClassVar[int]
    QUERYPLANINFO_FIELD_NUMBER: ClassVar[int]
    SCHEMA_FIELD_NUMBER: ClassVar[int]
    SPARKSCHEMA_FIELD_NUMBER: ClassVar[int]
    UNCAUGHTERROR_FIELD_NUMBER: ClassVar[int]
    VALIDATIONERROR_FIELD_NUMBER: ClassVar[int]
    listHiveResult: _hive_metastore__client_pb2.ListHiveResult
    multipleDataSourceSchemaResponse: GetMultipleDataSourceSchemasResponse
    multipleFeatureViewSchemaResponse: GetMultipleFeatureViewSchemasResponse
    queryPlanInfo: QueryPlanInfo
    schema: _schema__client_pb2.Schema
    sparkSchema: _spark_schema__client_pb2.SparkSchema
    uncaughtError: str
    validationError: str
    def __init__(self, uncaughtError: Optional[str] = ..., validationError: Optional[str] = ..., sparkSchema: Optional[Union[_spark_schema__client_pb2.SparkSchema, Mapping]] = ..., queryPlanInfo: Optional[Union[QueryPlanInfo, Mapping]] = ..., schema: Optional[Union[_schema__client_pb2.Schema, Mapping]] = ..., listHiveResult: Optional[Union[_hive_metastore__client_pb2.ListHiveResult, Mapping]] = ..., multipleFeatureViewSchemaResponse: Optional[Union[GetMultipleFeatureViewSchemasResponse, Mapping]] = ..., multipleDataSourceSchemaResponse: Optional[Union[GetMultipleDataSourceSchemasResponse, Mapping]] = ...) -> None: ...

class FeatureViewSchemaRequest(_message.Message):
    __slots__ = ["feature_view", "join_keys", "temporal_aggregate"]
    FEATURE_VIEW_FIELD_NUMBER: ClassVar[int]
    JOIN_KEYS_FIELD_NUMBER: ClassVar[int]
    TEMPORAL_AGGREGATE_FIELD_NUMBER: ClassVar[int]
    feature_view: _feature_view__client_pb2.FeatureViewArgs
    join_keys: _containers.RepeatedScalarFieldContainer[str]
    temporal_aggregate: _feature_view__client_pb2.TemporalAggregate
    def __init__(self, feature_view: Optional[Union[_feature_view__client_pb2.FeatureViewArgs, Mapping]] = ..., join_keys: Optional[Iterable[str]] = ..., temporal_aggregate: Optional[Union[_feature_view__client_pb2.TemporalAggregate, Mapping]] = ...) -> None: ...

class FeatureViewSchemaResponse(_message.Message):
    __slots__ = ["timestamp_key", "view_schema"]
    TIMESTAMP_KEY_FIELD_NUMBER: ClassVar[int]
    VIEW_SCHEMA_FIELD_NUMBER: ClassVar[int]
    timestamp_key: str
    view_schema: _schema__client_pb2.Schema
    def __init__(self, view_schema: Optional[Union[_schema__client_pb2.Schema, Mapping]] = ..., timestamp_key: Optional[str] = ...) -> None: ...

class GetBatchDataSourceFunctionSchema(_message.Message):
    __slots__ = ["function", "supports_time_filtering"]
    FUNCTION_FIELD_NUMBER: ClassVar[int]
    SUPPORTS_TIME_FILTERING_FIELD_NUMBER: ClassVar[int]
    function: _user_defined_function__client_pb2.UserDefinedFunction
    supports_time_filtering: bool
    def __init__(self, function: Optional[Union[_user_defined_function__client_pb2.UserDefinedFunction, Mapping]] = ..., supports_time_filtering: bool = ...) -> None: ...

class GetFeatureViewSchema(_message.Message):
    __slots__ = ["feature_view", "transformations", "virtual_data_sources"]
    FEATURE_VIEW_FIELD_NUMBER: ClassVar[int]
    TRANSFORMATIONS_FIELD_NUMBER: ClassVar[int]
    VIRTUAL_DATA_SOURCES_FIELD_NUMBER: ClassVar[int]
    feature_view: _feature_view__client_pb2.FeatureViewArgs
    transformations: _containers.RepeatedCompositeFieldContainer[_transformation__client_pb2.Transformation]
    virtual_data_sources: _containers.RepeatedCompositeFieldContainer[_virtual_data_source__client_pb2.VirtualDataSource]
    def __init__(self, virtual_data_sources: Optional[Iterable[Union[_virtual_data_source__client_pb2.VirtualDataSource, Mapping]]] = ..., transformations: Optional[Iterable[Union[_transformation__client_pb2.Transformation, Mapping]]] = ..., feature_view: Optional[Union[_feature_view__client_pb2.FeatureViewArgs, Mapping]] = ...) -> None: ...

class GetFileSourceSchema(_message.Message):
    __slots__ = ["convertToGlueFormat", "fileFormat", "rawBatchTranslator", "schemaOverride", "schemaUri", "timestampColumn", "timestampFormat", "uri"]
    CONVERTTOGLUEFORMAT_FIELD_NUMBER: ClassVar[int]
    FILEFORMAT_FIELD_NUMBER: ClassVar[int]
    RAWBATCHTRANSLATOR_FIELD_NUMBER: ClassVar[int]
    SCHEMAOVERRIDE_FIELD_NUMBER: ClassVar[int]
    SCHEMAURI_FIELD_NUMBER: ClassVar[int]
    TIMESTAMPCOLUMN_FIELD_NUMBER: ClassVar[int]
    TIMESTAMPFORMAT_FIELD_NUMBER: ClassVar[int]
    URI_FIELD_NUMBER: ClassVar[int]
    convertToGlueFormat: bool
    fileFormat: str
    rawBatchTranslator: _user_defined_function__client_pb2.UserDefinedFunction
    schemaOverride: _spark_schema__client_pb2.SparkSchema
    schemaUri: str
    timestampColumn: str
    timestampFormat: str
    uri: str
    def __init__(self, uri: Optional[str] = ..., fileFormat: Optional[str] = ..., convertToGlueFormat: bool = ..., rawBatchTranslator: Optional[Union[_user_defined_function__client_pb2.UserDefinedFunction, Mapping]] = ..., schemaUri: Optional[str] = ..., timestampColumn: Optional[str] = ..., timestampFormat: Optional[str] = ..., schemaOverride: Optional[Union[_spark_schema__client_pb2.SparkSchema, Mapping]] = ...) -> None: ...

class GetHiveTableSchema(_message.Message):
    __slots__ = ["database", "rawBatchTranslator", "table", "timestampColumn", "timestampFormat"]
    DATABASE_FIELD_NUMBER: ClassVar[int]
    RAWBATCHTRANSLATOR_FIELD_NUMBER: ClassVar[int]
    TABLE_FIELD_NUMBER: ClassVar[int]
    TIMESTAMPCOLUMN_FIELD_NUMBER: ClassVar[int]
    TIMESTAMPFORMAT_FIELD_NUMBER: ClassVar[int]
    database: str
    rawBatchTranslator: _user_defined_function__client_pb2.UserDefinedFunction
    table: str
    timestampColumn: str
    timestampFormat: str
    def __init__(self, database: Optional[str] = ..., table: Optional[str] = ..., timestampColumn: Optional[str] = ..., timestampFormat: Optional[str] = ..., rawBatchTranslator: Optional[Union[_user_defined_function__client_pb2.UserDefinedFunction, Mapping]] = ...) -> None: ...

class GetKafkaSourceSchema(_message.Message):
    __slots__ = ["rawStreamTranslator", "ssl_keystore_location", "ssl_keystore_password_secret_id"]
    RAWSTREAMTRANSLATOR_FIELD_NUMBER: ClassVar[int]
    SSL_KEYSTORE_LOCATION_FIELD_NUMBER: ClassVar[int]
    SSL_KEYSTORE_PASSWORD_SECRET_ID_FIELD_NUMBER: ClassVar[int]
    rawStreamTranslator: _user_defined_function__client_pb2.UserDefinedFunction
    ssl_keystore_location: str
    ssl_keystore_password_secret_id: str
    def __init__(self, rawStreamTranslator: Optional[Union[_user_defined_function__client_pb2.UserDefinedFunction, Mapping]] = ..., ssl_keystore_location: Optional[str] = ..., ssl_keystore_password_secret_id: Optional[str] = ...) -> None: ...

class GetKinesisSourceSchema(_message.Message):
    __slots__ = ["rawStreamTranslator", "streamName"]
    RAWSTREAMTRANSLATOR_FIELD_NUMBER: ClassVar[int]
    STREAMNAME_FIELD_NUMBER: ClassVar[int]
    rawStreamTranslator: _user_defined_function__client_pb2.UserDefinedFunction
    streamName: str
    def __init__(self, streamName: Optional[str] = ..., rawStreamTranslator: Optional[Union[_user_defined_function__client_pb2.UserDefinedFunction, Mapping]] = ...) -> None: ...

class GetMultipleDataSourceSchemasRequest(_message.Message):
    __slots__ = ["data_source_args"]
    DATA_SOURCE_ARGS_FIELD_NUMBER: ClassVar[int]
    data_source_args: _containers.RepeatedCompositeFieldContainer[_virtual_data_source__client_pb2.VirtualDataSourceArgs]
    def __init__(self, data_source_args: Optional[Iterable[Union[_virtual_data_source__client_pb2.VirtualDataSourceArgs, Mapping]]] = ...) -> None: ...

class GetMultipleDataSourceSchemasResponse(_message.Message):
    __slots__ = ["data_source_responses"]
    class DataSourceResponsesEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: SparkSchemas
        def __init__(self, key: Optional[str] = ..., value: Optional[Union[SparkSchemas, Mapping]] = ...) -> None: ...
    DATA_SOURCE_RESPONSES_FIELD_NUMBER: ClassVar[int]
    data_source_responses: _containers.MessageMap[str, SparkSchemas]
    def __init__(self, data_source_responses: Optional[Mapping[str, SparkSchemas]] = ...) -> None: ...

class GetMultipleFeatureViewSchemasRequest(_message.Message):
    __slots__ = ["feature_view_requests", "transformations", "virtual_data_sources"]
    FEATURE_VIEW_REQUESTS_FIELD_NUMBER: ClassVar[int]
    TRANSFORMATIONS_FIELD_NUMBER: ClassVar[int]
    VIRTUAL_DATA_SOURCES_FIELD_NUMBER: ClassVar[int]
    feature_view_requests: _containers.RepeatedCompositeFieldContainer[FeatureViewSchemaRequest]
    transformations: _containers.RepeatedCompositeFieldContainer[_transformation__client_pb2.Transformation]
    virtual_data_sources: _containers.RepeatedCompositeFieldContainer[_virtual_data_source__client_pb2.VirtualDataSource]
    def __init__(self, virtual_data_sources: Optional[Iterable[Union[_virtual_data_source__client_pb2.VirtualDataSource, Mapping]]] = ..., transformations: Optional[Iterable[Union[_transformation__client_pb2.Transformation, Mapping]]] = ..., feature_view_requests: Optional[Iterable[Union[FeatureViewSchemaRequest, Mapping]]] = ...) -> None: ...

class GetMultipleFeatureViewSchemasResponse(_message.Message):
    __slots__ = ["feature_view_responses"]
    class FeatureViewResponsesEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: ClassVar[int]
        VALUE_FIELD_NUMBER: ClassVar[int]
        key: str
        value: FeatureViewSchemaResponse
        def __init__(self, key: Optional[str] = ..., value: Optional[Union[FeatureViewSchemaResponse, Mapping]] = ...) -> None: ...
    FEATURE_VIEW_RESPONSES_FIELD_NUMBER: ClassVar[int]
    feature_view_responses: _containers.MessageMap[str, FeatureViewSchemaResponse]
    def __init__(self, feature_view_responses: Optional[Mapping[str, FeatureViewSchemaResponse]] = ...) -> None: ...

class GetQueryPlanInfoForFeatureViewPipeline(_message.Message):
    __slots__ = ["feature_view", "transformations", "virtual_data_sources"]
    FEATURE_VIEW_FIELD_NUMBER: ClassVar[int]
    TRANSFORMATIONS_FIELD_NUMBER: ClassVar[int]
    VIRTUAL_DATA_SOURCES_FIELD_NUMBER: ClassVar[int]
    feature_view: _feature_view__client_pb2.FeatureViewArgs
    transformations: _containers.RepeatedCompositeFieldContainer[_transformation__client_pb2.Transformation]
    virtual_data_sources: _containers.RepeatedCompositeFieldContainer[_virtual_data_source__client_pb2.VirtualDataSource]
    def __init__(self, virtual_data_sources: Optional[Iterable[Union[_virtual_data_source__client_pb2.VirtualDataSource, Mapping]]] = ..., transformations: Optional[Iterable[Union[_transformation__client_pb2.Transformation, Mapping]]] = ..., feature_view: Optional[Union[_feature_view__client_pb2.FeatureViewArgs, Mapping]] = ...) -> None: ...

class GetRedshiftTableSchema(_message.Message):
    __slots__ = ["endpoint", "query", "rawBatchTranslator", "table", "temp_s3_dir"]
    ENDPOINT_FIELD_NUMBER: ClassVar[int]
    QUERY_FIELD_NUMBER: ClassVar[int]
    RAWBATCHTRANSLATOR_FIELD_NUMBER: ClassVar[int]
    TABLE_FIELD_NUMBER: ClassVar[int]
    TEMP_S3_DIR_FIELD_NUMBER: ClassVar[int]
    endpoint: str
    query: str
    rawBatchTranslator: _user_defined_function__client_pb2.UserDefinedFunction
    table: str
    temp_s3_dir: str
    def __init__(self, endpoint: Optional[str] = ..., table: Optional[str] = ..., query: Optional[str] = ..., rawBatchTranslator: Optional[Union[_user_defined_function__client_pb2.UserDefinedFunction, Mapping]] = ..., temp_s3_dir: Optional[str] = ...) -> None: ...

class GetSnowflakeSchema(_message.Message):
    __slots__ = ["database", "post_processor", "query", "role", "schema", "table", "url", "warehouse"]
    DATABASE_FIELD_NUMBER: ClassVar[int]
    POST_PROCESSOR_FIELD_NUMBER: ClassVar[int]
    QUERY_FIELD_NUMBER: ClassVar[int]
    ROLE_FIELD_NUMBER: ClassVar[int]
    SCHEMA_FIELD_NUMBER: ClassVar[int]
    TABLE_FIELD_NUMBER: ClassVar[int]
    URL_FIELD_NUMBER: ClassVar[int]
    WAREHOUSE_FIELD_NUMBER: ClassVar[int]
    database: str
    post_processor: _user_defined_function__client_pb2.UserDefinedFunction
    query: str
    role: str
    schema: str
    table: str
    url: str
    warehouse: str
    def __init__(self, url: Optional[str] = ..., role: Optional[str] = ..., database: Optional[str] = ..., schema: Optional[str] = ..., warehouse: Optional[str] = ..., table: Optional[str] = ..., query: Optional[str] = ..., post_processor: Optional[Union[_user_defined_function__client_pb2.UserDefinedFunction, Mapping]] = ...) -> None: ...

class GetStreamDataSourceFunctionSchema(_message.Message):
    __slots__ = ["function"]
    FUNCTION_FIELD_NUMBER: ClassVar[int]
    function: _user_defined_function__client_pb2.UserDefinedFunction
    def __init__(self, function: Optional[Union[_user_defined_function__client_pb2.UserDefinedFunction, Mapping]] = ...) -> None: ...

class GetUnityTableSchema(_message.Message):
    __slots__ = ["catalog", "rawBatchTranslator", "schema", "table", "timestampColumn", "timestampFormat"]
    CATALOG_FIELD_NUMBER: ClassVar[int]
    RAWBATCHTRANSLATOR_FIELD_NUMBER: ClassVar[int]
    SCHEMA_FIELD_NUMBER: ClassVar[int]
    TABLE_FIELD_NUMBER: ClassVar[int]
    TIMESTAMPCOLUMN_FIELD_NUMBER: ClassVar[int]
    TIMESTAMPFORMAT_FIELD_NUMBER: ClassVar[int]
    catalog: str
    rawBatchTranslator: _user_defined_function__client_pb2.UserDefinedFunction
    schema: str
    table: str
    timestampColumn: str
    timestampFormat: str
    def __init__(self, catalog: Optional[str] = ..., schema: Optional[str] = ..., table: Optional[str] = ..., timestampColumn: Optional[str] = ..., timestampFormat: Optional[str] = ..., rawBatchTranslator: Optional[Union[_user_defined_function__client_pb2.UserDefinedFunction, Mapping]] = ...) -> None: ...

class ListHiveDatabases(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class ListHiveTableColumns(_message.Message):
    __slots__ = ["database", "table"]
    DATABASE_FIELD_NUMBER: ClassVar[int]
    TABLE_FIELD_NUMBER: ClassVar[int]
    database: str
    table: str
    def __init__(self, database: Optional[str] = ..., table: Optional[str] = ...) -> None: ...

class ListHiveTables(_message.Message):
    __slots__ = ["database"]
    DATABASE_FIELD_NUMBER: ClassVar[int]
    database: str
    def __init__(self, database: Optional[str] = ...) -> None: ...

class QueryPlanInfo(_message.Message):
    __slots__ = ["has_aggregations", "has_joins"]
    HAS_AGGREGATIONS_FIELD_NUMBER: ClassVar[int]
    HAS_JOINS_FIELD_NUMBER: ClassVar[int]
    has_aggregations: bool
    has_joins: bool
    def __init__(self, has_joins: bool = ..., has_aggregations: bool = ...) -> None: ...

class SparkSchemas(_message.Message):
    __slots__ = ["batchSchema", "streamSchema"]
    BATCHSCHEMA_FIELD_NUMBER: ClassVar[int]
    STREAMSCHEMA_FIELD_NUMBER: ClassVar[int]
    batchSchema: _spark_schema__client_pb2.SparkSchema
    streamSchema: _spark_schema__client_pb2.SparkSchema
    def __init__(self, batchSchema: Optional[Union[_spark_schema__client_pb2.SparkSchema, Mapping]] = ..., streamSchema: Optional[Union[_spark_schema__client_pb2.SparkSchema, Mapping]] = ...) -> None: ...



from dataclasses import dataclass
from typing import Dict
from typing import Optional

from tecton_core.data_types import DataType


@dataclass
class FeatureMetadataSpec:
    name: str
    dtype: DataType
    description: Optional[str]
    tags: Optional[Dict[str, str]]
